import mock
import uuid

from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.errors import ERRORS
from kinto.core.testing import unittest, FormattedErrorMixin


from ..support import BaseWebTest


MINIMALIST_RECORD = {'name': 'Champignon'}


class UserResourcePermissionTest(BaseWebTest, unittest.TestCase):
    authorization_policy = 'kinto.core.authorization.AuthorizationPolicy'

    def test_views_require_authentication(self):
        self.app.get(self.collection_url, status=401)
        body = {'data': MINIMALIST_RECORD}
        self.app.post_json(self.collection_url, body, status=401)
        record_url = self.get_item_url('abc')
        self.app.get(record_url, status=401)
        self.app.put_json(record_url, body, status=401)
        self.app.patch_json(record_url, body, status=401)
        self.app.delete(record_url, status=401)

    def test_collection_operations_are_authorized_if_authenticated(self):
        body = {'data': MINIMALIST_RECORD}
        self.app.get(self.collection_url, headers=self.headers, status=200)
        self.app.post_json(self.collection_url, body,
                           headers=self.headers, status=201)
        self.app.delete(self.collection_url, headers=self.headers, status=200)

    def test_record_operations_are_authorized_if_authenticated(self):
        body = {'data': MINIMALIST_RECORD}
        resp = self.app.post_json(self.collection_url, body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])
        unknown_url = self.get_item_url(uuid.uuid4())

        self.app.get(record_url, headers=self.headers, status=200)
        self.app.patch_json(record_url, body, headers=self.headers, status=200)
        self.app.delete(record_url, headers=self.headers, status=200)
        self.app.put_json(record_url, body, headers=self.headers, status=201)
        self.app.put_json(unknown_url, body, headers=self.headers, status=201)


class AuthzAuthnTest(BaseWebTest, unittest.TestCase):
    authorization_policy = 'kinto.core.authorization.AuthorizationPolicy'
    # Shareable resource.
    collection_url = '/toadstools'

    def add_permission(self, object_id, permission, principal=None):
        if not principal:
            principal = self.principal
        self.permission.add_principal_to_ace(object_id, permission, principal)


class ShareableResourcePermissionTest(AuthzAuthnTest):
    def setUp(self):
        self.add_permission(self.collection_url, 'toadstool:create')

    def test_permissions_are_associated_to_object_uri_without_prefix(self):
        body = {'data': MINIMALIST_RECORD,
                'permissions': {'read': ['group:readers']}}
        resp = self.app.post_json(self.collection_url, body,
                                  headers=self.headers)
        object_uri = self.get_item_url(resp.json['data']['id'])
        backend = self.permission
        stored_perms = backend.get_object_permission_principals(object_uri,
                                                                'read')
        self.assertEqual(stored_perms, {'group:readers'})

    def test_permissions_are_not_modified_if_not_specified(self):
        body = {'data': MINIMALIST_RECORD,
                'permissions': {'read': ['group:readers']}}
        resp = self.app.post_json(self.collection_url, body,
                                  headers=self.headers)
        object_uri = self.get_item_url(resp.json['data']['id'])
        body.pop('permissions')

        self.add_permission(object_uri, 'write')
        resp = self.app.put_json(object_uri, body, headers=self.headers)
        self.assertEqual(resp.json['permissions']['read'], ['group:readers'])

    def test_data_is_always_required_when_schema_has_required_fields(self):
        body = {'data': MINIMALIST_RECORD,
                'permissions': {'read': ['group:readers']}}
        resp = self.app.post_json(self.collection_url, body,
                                  headers=self.headers)
        object_uri = self.get_item_url(resp.json['data']['id'])

        body.pop('data')
        resp = self.app.put_json(object_uri, body, headers=self.headers,
                                 status=400)

    def test_data_is_not_required_when_schema_has_no_required_fields(self):
        self.add_permission('/psilos', 'psilo:create')
        body = {'data': MINIMALIST_RECORD,
                'permissions': {'read': ['group:readers']}}
        resp = self.app.post_json('/psilos', body,
                                  headers=self.headers)
        object_uri = '/psilos/{}'.format(resp.json['data']['id'])

        body.pop('data')
        resp = self.app.put_json(object_uri, body, headers=self.headers)
        self.assertEqual(resp.json['data']['name'], MINIMALIST_RECORD['name'])

    def test_data_are_not_modified_if_not_specified_on_schemaless(self):
        self.add_permission('/spores', 'spore:create')
        body = {'data': MINIMALIST_RECORD,
                'permissions': {'read': ['group:readers']}}
        resp = self.app.post_json('/spores', body,
                                  headers=self.headers)
        object_uri = '/spores/{}'.format(resp.json['data']['id'])
        self.add_permission(object_uri, 'write')

        body.pop('data')
        resp = self.app.put_json(object_uri, body, headers=self.headers)

        self.assertEqual(resp.json['data']['name'], MINIMALIST_RECORD['name'])

    def test_permissions_can_be_modified_using_patch(self):
        body = {'data': MINIMALIST_RECORD,
                'permissions': {'read': ['group:readers']}}
        resp = self.app.post_json(self.collection_url, body,
                                  headers=self.headers)
        body = {'permissions': {'read': ['fxa:user']}}
        uri = self.get_item_url(resp.json['data']['id'])
        resp = self.app.patch_json(uri, body, headers=self.headers)
        principals = resp.json['permissions']['read']
        self.assertEqual(sorted(principals), ['fxa:user'])

    def test_unknown_permissions_are_not_accepted(self):
        self.maxDiff = None
        body = {'data': MINIMALIST_RECORD,
                'permissions': {'read': ['group:readers'],
                                'unknown': ['jacques']}}
        resp = self.app.post_json(self.collection_url, body,
                                  headers=self.headers, status=400)
        self.assertEqual(
            resp.json['message'],
            'permissions in body: "unknown" is not one of read, write')


class CollectionAuthzGrantedTest(AuthzAuthnTest):
    def test_collection_get_is_granted_when_authorized(self):
        self.add_permission(self.collection_url, 'read')
        self.app.get(self.collection_url, headers=self.headers, status=200)

    def test_collection_post_is_granted_when_authorized(self):
        self.add_permission(self.collection_url, 'toadstool:create')
        self.app.post_json(self.collection_url, {'data': MINIMALIST_RECORD},
                           headers=self.headers, status=201)

    def test_collection_delete_is_granted_when_authorized(self):
        self.add_permission(self.collection_url, 'write')
        self.app.delete(self.collection_url, headers=self.headers, status=200)


class CollectionAuthzDeniedTest(AuthzAuthnTest):
    def test_views_require_authentication(self):
        self.app.get(self.collection_url, status=401)

        body = {'data': MINIMALIST_RECORD}
        self.app.post_json(self.collection_url, body, status=401)

    def test_collection_get_is_denied_when_not_authorized(self):
        self.app.get(self.collection_url, headers=self.headers, status=403)

    def test_collection_post_is_denied_when_not_authorized(self):
        self.app.post_json(self.collection_url, {'data': MINIMALIST_RECORD},
                           headers=self.headers, status=403)

    def test_collection_delete_is_denied_when_not_authorized(self):
        self.app.delete(self.collection_url, headers=self.headers, status=403)


class RecordAuthzGrantedTest(AuthzAuthnTest):
    def setUp(self):
        super().setUp()
        self.add_permission(self.collection_url, 'toadstool:create')

        resp = self.app.post_json(self.collection_url,
                                  {'data': MINIMALIST_RECORD},
                                  headers=self.headers)
        self.record = resp.json['data']
        self.record_url = self.get_item_url()
        self.unknown_record_url = self.get_item_url(uuid.uuid4())

    def test_record_get_is_granted_when_authorized(self):
        self.add_permission(self.record_url, 'read')
        self.app.get(self.record_url, headers=self.headers, status=200)

    def test_record_patch_is_granted_when_authorized(self):
        self.add_permission(self.record_url, 'write')
        self.app.patch_json(self.record_url, {'data': MINIMALIST_RECORD},
                            headers=self.headers, status=200)

    def test_record_delete_is_granted_when_authorized(self):
        self.add_permission(self.record_url, 'write')
        self.app.delete(self.record_url, headers=self.headers, status=200)

    def test_record_put_on_existing_record_is_granted_when_authorized(self):
        self.add_permission(self.record_url, 'write')
        self.app.put_json(self.record_url, {'data': MINIMALIST_RECORD},
                          headers=self.headers, status=200)

    def test_record_put_on_unexisting_record_is_granted_when_authorized(self):
        self.add_permission(self.collection_url, 'toadstool:create')
        self.app.put_json(self.unknown_record_url, {'data': MINIMALIST_RECORD},
                          headers=self.headers, status=201)


class RecordAuthzDeniedTest(AuthzAuthnTest):
    def setUp(self):
        super().setUp()
        # Add permission to create a sample record.
        self.add_permission(self.collection_url, 'toadstool:create')
        resp = self.app.post_json(self.collection_url,
                                  {'data': MINIMALIST_RECORD},
                                  headers=self.headers)
        self.record = resp.json['data']
        self.record_url = self.get_item_url()
        self.unknown_record_url = self.get_item_url(uuid.uuid4())
        # Remove every permissions.
        self.permission.flush()

    def test_views_require_authentication(self):
        url = self.get_item_url('abc')
        self.app.get(url, status=401)
        self.app.put_json(url, {'data': MINIMALIST_RECORD}, status=401)
        self.app.patch_json(url, {'data': MINIMALIST_RECORD}, status=401)
        self.app.delete(url, status=401)

    def test_record_get_is_denied_when_not_authorized(self):
        self.app.get(self.record_url, headers=self.headers, status=403)

    def test_record_patch_is_denied_when_not_authorized(self):
        self.app.patch_json(self.record_url, {'data': MINIMALIST_RECORD},
                            headers=self.headers, status=403)

    def test_record_put_is_denied_when_not_authorized(self):
        self.app.put_json(self.record_url, {'data': MINIMALIST_RECORD},
                          headers=self.headers, status=403)

    def test_record_delete_is_denied_when_not_authorized(self):
        self.app.delete(self.record_url, headers=self.headers, status=403)

    def test_record_put_on_unexisting_record_is_rejected_if_write_perm(self):
        object_id = self.collection_url
        self.permission.remove_principal_from_ace(
            object_id, 'toadstool:create', self.principal)  # Added in setUp.

        self.permission.add_principal_to_ace(
            object_id, 'write', self.principal)
        self.app.put_json(self.unknown_record_url, {'data': MINIMALIST_RECORD},
                          headers=self.headers, status=403)


class RecordAuthzGrantedOnCollectionTest(AuthzAuthnTest):
    def setUp(self):
        super().setUp()
        self.add_permission(self.collection_url, 'toadstool:create')

        self.guest_headers = {**self.headers, 'Authorization': "Basic bmF0aW06"}
        resp = self.app.get('/', headers=self.guest_headers)
        self.guest_id = resp.json['user']['id']

        body = {
            'data': MINIMALIST_RECORD,
            'permissions': {
                'write': [self.guest_id],
                'read': [self.guest_id]
            }
        }
        resp = self.app.post_json(self.collection_url, body,
                                  headers=self.headers)
        self.record = resp.json['data']
        self.record_url = self.get_item_url()

        # Add another private record
        resp = self.app.post_json(self.collection_url,
                                  {'data': MINIMALIST_RECORD,
                                   'permissions': {"read": [self.principal]}},
                                  headers=self.headers)

    def test_guest_can_access_the_record(self):
        self.app.get(self.record_url, headers=self.guest_headers, status=200)

    def test_guest_can_see_the_record_in_the_list_of_records(self):
        resp = self.app.get(self.collection_url, headers=self.guest_headers,
                            status=200)
        self.assertEqual(len(resp.json['data']), 1)

    def test_guest_can_remove_its_records_from_the_list_of_records(self):
        resp = self.app.delete(self.collection_url, headers=self.guest_headers,
                               status=200)
        self.assertEqual(len(resp.json['data']), 1)
        resp = self.app.get(self.collection_url, headers=self.headers,
                            status=200)
        self.assertEqual(len(resp.json['data']), 1)


class StrictSchemaTest(BaseWebTest, unittest.TestCase):
    collection_url = '/moistures'

    def test_accept_empty_body(self):
        resp = self.app.post(self.collection_url,
                             headers=self.headers)
        self.assertIn('id', resp.json['data'])
        resp = self.app.put(self.get_item_url(uuid.uuid4()),
                            headers=self.headers)
        self.assertIn('id', resp.json['data'])

    def test_data_can_be_specified(self):
        resp = self.app.post_json(self.collection_url,
                                  {'data': {}},
                                  headers=self.headers)
        self.assertIn('id', resp.json['data'])

    def test_data_fields_are_ignored(self):
        resp = self.app.post_json(self.collection_url,
                                  {'data': {'icq': '9427392'}},
                                  headers=self.headers)
        self.assertNotIn('icq', resp.json['data'])


class OptionalSchemaTest(BaseWebTest, unittest.TestCase):
    collection_url = '/psilos'

    def test_accept_empty_body(self):
        resp = self.app.post(self.collection_url,
                             headers=self.headers)
        self.assertIn('id', resp.json['data'])
        resp = self.app.put(self.get_item_url(uuid.uuid4()),
                            headers=self.headers)
        self.assertIn('id', resp.json['data'])

    def test_data_can_be_specified(self):
        resp = self.app.post_json(self.collection_url,
                                  {'data': {}},
                                  headers=self.headers)
        self.assertIn('id', resp.json['data'])

    def test_known_fields_are_saved(self):
        resp = self.app.post_json(self.collection_url,
                                  {'data': {'edible': False}},
                                  headers=self.headers)
        self.assertIn('edible', resp.json['data'])


class InvalidRecordTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        body = {'data': MINIMALIST_RECORD}
        resp = self.app.post_json(self.collection_url,
                                  body,
                                  headers=self.headers)
        self.record = resp.json['data']

        self.invalid_record = {'data': {'name': 42}}

    def test_invalid_record_returns_json_formatted_error(self):
        resp = self.app.post_json(self.collection_url,
                                  self.invalid_record,
                                  headers=self.headers,
                                  status=400)
        # XXX: weird resp.json['message']
        self.assertDictEqual(resp.json, {
            'errno': ERRORS.INVALID_PARAMETERS.value,
            'message': "data.name in body: 42 is not a string: {'name': ''}",
            'code': 400,
            'error': 'Invalid parameters',
            'details': [{'description': "42 is not a string: {'name': ''}",
                         'location': 'body',
                         'name': 'data.name'}]})

    def test_empty_body_returns_400(self):
        resp = self.app.post(self.collection_url,
                             '',
                             headers=self.headers,
                             status=400)
        self.assertEqual(resp.json['message'], 'data in body: Required')

    def test_unknown_attribute_returns_400(self):
        resp = self.app.post(self.collection_url,
                             '{"data": {"name": "ML"}, "datta": {}}',
                             headers=self.headers,
                             status=400)
        self.assertIn('Unrecognized keys in mapping', resp.json['message'])
        self.assertIn('datta', resp.json['message'])

    def test_create_invalid_record_returns_400(self):
        self.app.post_json(self.collection_url,
                           self.invalid_record,
                           headers=self.headers,
                           status=400)

    def test_modify_with_invalid_record_returns_400(self):
        self.app.patch_json(self.get_item_url(),
                            self.invalid_record,
                            headers=self.headers,
                            status=400)

    def test_modify_with_unknown_attribute_returns_400(self):
        self.app.patch_json(self.get_item_url(),
                            {"datta": {}},
                            headers=self.headers,
                            status=400)

    def test_replace_with_invalid_record_returns_400(self):
        self.app.put_json(self.get_item_url(),
                          self.invalid_record,
                          headers=self.headers,
                          status=400)

    def test_id_is_validated_on_post(self):
        record = {**MINIMALIST_RECORD, 'id': 3.14}
        self.app.post_json(self.collection_url,
                           {'data': record},
                           headers=self.headers,
                           status=400)

        with mock.patch.object(self.app.app.registry.id_generators[''],
                               'match', return_value=True):
            self.app.post_json(self.collection_url,
                               {'data': record},
                               headers=self.headers,
                               status=400)

    def test_id_is_preserved_on_post(self):
        record = {**MINIMALIST_RECORD, 'id': '472be9ec-26fe-461b-8282-9c4e4b207ab3'}
        resp = self.app.post_json(self.collection_url,
                                  {'data': record},
                                  headers=self.headers)
        self.assertEqual(resp.json['data']['id'], record['id'])

    def test_200_is_returned_if_id_matches_existing_record(self):
        record = {**MINIMALIST_RECORD, 'id': self.record['id']}
        self.app.post_json(self.collection_url,
                           {'data': record},
                           headers=self.headers,
                           status=200)

    def test_invalid_accept_header_on_collections_returns_406(self):
        headers = {**self.headers, 'Accept': 'text/plain'}
        resp = self.app.post(self.collection_url, '', headers=headers, status=406)
        self.assertEqual(resp.json['code'], 406)
        message = "Accept header should be one of ['application/json']"
        self.assertEqual(resp.json['message'], message)

    def test_invalid_content_type_header_on_collections_returns_415(self):
        headers = {**self.headers, 'Content-Type': 'text/plain'}
        resp = self.app.post(self.collection_url, '', headers=headers, status=415)
        self.assertEqual(resp.json['code'], 415)
        message = "Content-Type header should be one of ['application/json']"
        self.assertEqual(resp.json['message'], message)

    def test_invalid_accept_header_on_record_returns_406(self):
        headers = {**self.headers, 'Accept': 'text/plain'}
        resp = self.app.get(self.get_item_url(), headers=headers, status=406)
        self.assertEqual(resp.json['code'], 406)
        message = "Accept header should be one of ['application/json']"
        self.assertEqual(resp.json['message'], message)

    def test_invalid_content_type_header_on_record_returns_415(self):
        headers = {**self.headers, 'Content-Type': 'text/plain'}
        resp = self.app.patch_json(self.get_item_url(), '', headers=headers, status=415)
        self.assertEqual(resp.json['code'], 415)
        messages = (
            "Content-Type header should be one of [",
            "'application/json-patch+json'",
            ", ",
            "'application/json'",
            ", ",
            "'application/merge-patch+json'",
            "]")
        for message in messages:
            self.assertIn(message, resp.json['message'])
        self.assertEquals(len("".join(messages)), len(resp.json['message']))


class IgnoredFieldsTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        body = {'data': MINIMALIST_RECORD}
        resp = self.app.post_json(self.collection_url,
                                  body,
                                  headers=self.headers)
        self.record = resp.json['data']

    def test_last_modified_is_not_validated_and_overwritten(self):
        record = {**MINIMALIST_RECORD, 'last_modified': 'abc'}
        resp = self.app.post_json(self.collection_url,
                                  {'data': record},
                                  headers=self.headers)
        self.assertNotEqual(resp.json['data']['last_modified'], 'abc')

    def test_modify_works_with_invalid_last_modified(self):
        body = {'data': {'last_modified': 'abc'}}
        resp = self.app.patch_json(self.get_item_url(),
                                   body,
                                   headers=self.headers)
        self.assertNotEqual(resp.json['data']['last_modified'], 'abc')

    def test_replace_works_with_invalid_last_modified(self):
        record = {**MINIMALIST_RECORD, 'last_modified': 'abc'}
        resp = self.app.put_json(self.get_item_url(),
                                 {'data': record},
                                 headers=self.headers)
        self.assertNotEqual(resp.json['data']['last_modified'], 'abc')


class InvalidBodyTest(BaseWebTest, unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.invalid_body = "{'foo>}"

    def setUp(self):
        super().setUp()
        body = {'data': MINIMALIST_RECORD}
        resp = self.app.post_json(self.collection_url,
                                  body,
                                  headers=self.headers)
        self.record = resp.json['data']

    def test_invalid_body_returns_json_formatted_error(self):
        self.maxDiff = None
        resp = self.app.post(self.collection_url,
                             self.invalid_body,
                             headers=self.headers,
                             status=400)
        error_msg = ("Invalid JSON: Expecting property name enclosed in "
                     "double quotes: line 1 column 2 (char 1)")
        self.assertDictEqual(resp.json, {
            'errno': ERRORS.INVALID_PARAMETERS.value,
            'message': error_msg,
            'code': 400,
            'error': 'Invalid parameters',
            'details': [
                {'description': error_msg,
                 'location': 'body',
                 'name': ''},
                {'description': 'Required',
                 'location': 'body',
                 'name': ''}]})

    def test_create_invalid_body_returns_400(self):
        self.app.post(self.collection_url,
                      self.invalid_body,
                      headers=self.headers,
                      status=400)

    def test_modify_with_invalid_body_returns_400(self):
        self.app.patch(self.get_item_url(),
                       self.invalid_body,
                       headers=self.headers,
                       status=400)

    def test_replace_with_invalid_body_returns_400(self):
        self.app.put(self.get_item_url(),
                     self.invalid_body,
                     headers=self.headers,
                     status=400)

    def test_invalid_uft8_returns_400(self):
        body = '{"foo": "\\u0d1"}'
        resp = self.app.post(self.collection_url,
                             body,
                             headers=self.headers,
                             status=400)
        self.assertIn('Invalid \\uXXXX escape: line 1', resp.json['message'])

    def test_modify_with_invalid_uft8_returns_400(self):
        body = '{"foo": "\\u0d1"}'
        resp = self.app.patch(self.get_item_url(),
                              body,
                              headers=self.headers,
                              status=400)
        self.assertIn('Invalid \\uXXXX escape', resp.json['message'])

    def test_modify_with_empty_body_returns_400(self):
        self.app.patch(self.get_item_url(),
                       headers=self.headers,
                       status=400)

    def test_modify_shareable_resource_with_empty_body_returns_400(self):
        body = {'data': MINIMALIST_RECORD}
        resp = self.app.post_json('/toadstools',
                                  body,
                                  headers=self.headers)
        record = resp.json['data']
        item_url = '/toadstools/{}'.format(record['id'])
        self.app.patch(item_url,
                       headers=self.headers,
                       status=400)


class InvalidPermissionsTest(BaseWebTest, unittest.TestCase):
    collection_url = '/toadstools'

    def setUp(self):
        super().setUp()
        body = {'data': MINIMALIST_RECORD}
        resp = self.app.post_json(self.collection_url,
                                  body,
                                  headers=self.headers)
        self.record = resp.json['data']
        self.invalid_body = {'data': MINIMALIST_RECORD,
                             'permissions': {'read': 'book'}}  # book not list

    def test_permissions_are_not_accepted_on_normal_resources(self):
        body = {'data': MINIMALIST_RECORD,
                'permissions': {'read': ['book']}}
        resp = self.app.post_json('/mushrooms', body, headers=self.headers,
                                  status=400)
        self.assertIn('Unrecognized keys in mapping', resp.json['message'])
        self.assertIn('permissions', resp.json['message'])

    def test_create_invalid_body_returns_400(self):
        self.app.post_json(self.collection_url,
                           self.invalid_body,
                           headers=self.headers,
                           status=400)

    def test_modify_with_invalid_permissions_returns_400(self):
        self.app.patch_json(self.get_item_url(),
                            self.invalid_body,
                            headers=self.headers,
                            status=400)

    def test_invalid_body_returns_json_formatted_error(self):
        resp = self.app.post_json(self.collection_url,
                                  self.invalid_body,
                                  headers=self.headers,
                                  status=400)
        self.assertDictEqual(resp.json, {
            'errno': ERRORS.INVALID_PARAMETERS.value,
            'message': 'permissions.read in body: "book" is not iterable',
            'code': 400,
            'error': 'Invalid parameters',
            'details': [
                {'description': '"book" is not iterable',
                 'location': 'body',
                 'name': 'permissions.read'}]})


class CacheControlTest(BaseWebTest, unittest.TestCase):
    collection_url = '/toadstools'

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['toadstool_cache_expires_seconds'] = 3600
        settings['toadstool_read_principals'] = 'system.Everyone'
        settings['psilo_cache_expires_seconds'] = 0
        settings['moisture_read_principals'] = 'system.Everyone'
        return settings

    def test_cache_control_headers_are_set_if_anonymous(self):
        resp = self.app.get(self.collection_url)
        self.assertIn('Expires', resp.headers)
        self.assertIn('Cache-Control', resp.headers)

    def test_cache_control_headers_are_not_set_if_authenticated(self):
        resp = self.app.get(self.collection_url, headers=self.headers)
        self.assertIn('no-cache', resp.headers['Cache-Control'])
        self.assertIn('no-store', resp.headers['Cache-Control'])
        self.assertNotIn('Expires', resp.headers)

    def test_cache_control_headers_set_no_cache_if_zero(self):
        resp = self.app.get('/psilos')
        self.assertIn('Cache-Control', resp.headers)
        # Check that not set by Pyramid.cache_expires()
        self.assertNotIn('Expires', resp.headers)

    def test_cache_control_provides_no_cache_by_default(self):
        resp = self.app.get('/moistures')
        self.assertIn('no-cache', resp.headers['Cache-Control'])
        self.assertNotIn('Expires', resp.headers)


class StorageErrorTest(BaseWebTest, unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error = storage_exceptions.BackendError(ValueError())
        self.storage_error_patcher = mock.patch(
            'kinto.core.storage.memory.Storage.create',
            side_effect=self.error)

    def test_backend_errors_are_served_as_503(self):
        body = {'data': MINIMALIST_RECORD}
        with self.storage_error_patcher:
            self.app.post_json(self.collection_url,
                               body,
                               headers=self.headers,
                               status=503)

    def test_backend_errors_original_error_is_logged(self):
        body = {'data': MINIMALIST_RECORD}
        with mock.patch('kinto.core.views.errors.logger.critical') as mocked:
            with self.storage_error_patcher:
                self.app.post_json(self.collection_url,
                                   body,
                                   headers=self.headers,
                                   status=503)
                self.assertTrue(mocked.called)
                self.assertEqual(type(mocked.call_args[0][0]), ValueError)


class PaginationNextURLTest(BaseWebTest, unittest.TestCase):
    """Extra tests for `tests.core.resource.test_pagination`
    """

    def setUp(self):
        super().setUp()
        body = {'data': MINIMALIST_RECORD}
        self.app.post_json(self.collection_url,
                           body,
                           headers=self.headers)
        self.app.post_json(self.collection_url,
                           body,
                           headers=self.headers)

    def test_next_page_url_has_got_port_number_if_different_than_80(self):
        resp = self.app.get(self.collection_url + '?_limit=1',
                            extra_environ={'HTTP_HOST': 'localhost:8000'},
                            headers=self.headers)
        self.assertIn(':8000', resp.headers['Next-Page'])

    def test_next_page_url_has_not_port_number_if_80(self):
        resp = self.app.get(self.collection_url + '?_limit=1',
                            extra_environ={'HTTP_HOST': 'localhost:80'},
                            headers=self.headers)
        self.assertNotIn(':80', resp.headers['Next-Page'])

    def test_next_page_url_relies_on_pyramid_url_system(self):
        resp = self.app.get(self.collection_url + '?_limit=1',
                            extra_environ={'wsgi.url_scheme': 'https'},
                            headers=self.headers)
        self.assertIn('https://', resp.headers['Next-Page'])

    def test_next_page_url_relies_on_headers_information(self):
        headers = {**self.headers, 'Host': 'https://server.name:443'}
        resp = self.app.get(self.collection_url + '?_limit=1',
                            headers=headers)
        self.assertIn('https://server.name:443', resp.headers['Next-Page'])


class SchemaLessPartialResponseTest(BaseWebTest, unittest.TestCase):
    """Extra tests for :mod:`tests.core.resource.test_partial_response`
    """
    collection_url = '/spores'

    def setUp(self):
        super().setUp()
        body = {'data': {'size': 42, 'category': 'some-cat', 'owner': 'loco'}}
        resp = self.app.post_json(self.collection_url,
                                  body,
                                  headers=self.headers)
        self.record = resp.json

    def test_unspecified_fields_are_excluded(self):
        resp = self.app.get(self.collection_url + '?_fields=size,category')
        result = resp.json['data'][0]
        self.assertNotIn('owner', result)

    def test_specified_fields_are_included(self):
        resp = self.app.get(self.collection_url + '?_fields=size,category')
        result = resp.json['data'][0]
        self.assertIn('size', result)
        self.assertIn('category', result)

    def test_unknown_fields_are_ignored(self):
        resp = self.app.get(self.collection_url + '?_fields=nationality')
        result = resp.json['data'][0]
        self.assertNotIn('nationality', result)


class UnicodeDecodeErrorTest(BaseWebTest, FormattedErrorMixin, unittest.TestCase):
    collection_url = '/spores'

    def test_wrong_filter_encoding_raise_a_400_bad_request(self):
        resp = self.app.get(self.collection_url + '?foo\xe2\xfc\xa7bar',
                            headers=self.headers, status=400)
        self.assertFormattedError(resp, 400, ERRORS.INVALID_PARAMETERS, "Bad Request",
                                  "A request with an incorrect encoding in the querystring was"
                                  "received. Please make sure your requests are encoded in UTF-8")



import mock
import uuid

from kinto.core import authentication
from kinto.core import utils
from kinto.core.testing import DummyRequest, unittest

from .support import BaseWebTest


class AuthenticationPoliciesTest(BaseWebTest, unittest.TestCase):

    def test_basic_auth_is_accepted_by_default(self):
        self.app.get(self.collection_url, headers=self.headers, status=200)
        # Check that the capability is exposed on the homepage.
        resp = self.app.get('/')
        assert 'basicauth' in resp.json['capabilities']

    def test_basic_auth_is_accepted_if_enabled_in_settings(self):
        app = self.make_app({'multiauth.policies': 'basicauth'})
        app.get(self.collection_url, headers=self.headers, status=200)
        # Check that the capability is exposed on the homepage.
        resp = app.get('/')
        assert 'basicauth' in resp.json['capabilities']

    def test_basic_auth_is_declined_if_disabled_in_settings(self):
        app = self.make_app({
            'multiauth.policies': 'dummy',
            'multiauth.policy.dummy.use': ('pyramid.authentication.'
                                           'RepozeWho1AuthenticationPolicy')})
        app.get(self.collection_url, headers=self.headers, status=401)
        # Check that the capability is exposed on the homepage.
        resp = app.get('/')
        assert 'basicauth' not in resp.json['capabilities']

    def test_views_are_forbidden_if_unknown_auth_method(self):
        app = self.make_app({'multiauth.policies': 'basicauth'})
        self.headers['Authorization'] = 'Carrier'
        app.get(self.collection_url, headers=self.headers, status=401)
        self.headers['Authorization'] = 'Carrier pigeon'
        app.get(self.collection_url, headers=self.headers, status=401)

    def test_principals_are_fetched_from_permission_backend(self):
        patch = mock.patch(('tests.core.support.'
                            'AllowAuthorizationPolicy.permits'))
        self.addCleanup(patch.stop)
        mocked = patch.start()

        self.permission.add_user_principal(self.principal, 'group:admin')
        self.app.get(self.collection_url, headers=self.headers)

        _, principals, _ = mocked.call_args[0]
        self.assertIn('group:admin', principals)

    def test_user_principals_are_cached_per_user(self):
        patch = mock.patch.object(self.permission, 'get_user_principals',
                                  wraps=self.permission.get_user_principals)
        self.addCleanup(patch.stop)
        mocked = patch.start()
        batch = {
            "defaults": {
                "headers": self.headers,
                "path": "/mushrooms"
            },
            "requests": [
                {},
                {},
                {},
                {"headers": {"Authorization": "Basic Ym9iOg=="}},
                {"headers": {"Authorization": "Basic bWF0Og=="}},
            ]
        }
        self.app.post_json('/batch', batch)
        self.assertEqual(mocked.call_count, 3)

    def test_basicauth_hash_is_computed_only_once(self):
        # hmac_digest() is used in Basic Authentication only.
        patch = mock.patch('kinto.core.utils.hmac_digest', return_value='abcdef')
        self.addCleanup(patch.stop)
        mocked = patch.start()
        body = {"data": {"name": "haha"}}
        record_url = self.get_item_url(uuid.uuid4())
        self.app.put_json(record_url, body, headers=self.headers)
        self.assertEqual(mocked.call_count, 1)


class BasicAuthenticationPolicyTest(unittest.TestCase):
    def setUp(self):
        self.policy = authentication.BasicAuthAuthenticationPolicy()
        self.request = DummyRequest()
        self.request.headers['Authorization'] = 'Basic bWF0Og=='

    @mock.patch('kinto.core.utils.hmac_digest')
    def test_userid_is_hashed(self, mocked):
        mocked.return_value = 'yeah'
        user_id = self.policy.unauthenticated_userid(self.request)
        self.assertIn('yeah', user_id)

    def test_userid_is_built_using_password(self):
        auth_password = utils.encode64('user:secret1', encoding='ascii')
        self.request.headers['Authorization'] = 'Basic {}'.format(auth_password)
        user_id1 = self.policy.unauthenticated_userid(self.request)

        auth_password = utils.encode64('user:secret2', encoding='ascii')
        self.request.headers['Authorization'] = 'Basic {}'.format(auth_password)
        user_id2 = self.policy.unauthenticated_userid(self.request)

        self.assertNotEqual(user_id1, user_id2)

    def test_views_are_forbidden_if_basic_is_wrong(self):
        self.request.headers['Authorization'] = 'Basic abc'
        user_id = self.policy.unauthenticated_userid(self.request)
        self.assertIsNone(user_id)

    def test_returns_none_if_username_is_empty(self):
        auth_password = utils.encode64(':secret', encoding='ascii')
        self.request.headers['Authorization'] = 'Basic {}'.format(auth_password)
        user_id = self.policy.unauthenticated_userid(self.request)
        self.assertIsNone(user_id)

    def test_providing_empty_password_is_supported(self):
        auth_password = utils.encode64('secret:', encoding='ascii')
        self.request.headers['Authorization'] = 'Basic {}'.format(auth_password)
        user_id = self.policy.unauthenticated_userid(self.request)
        self.assertIsNotNone(user_id)



import mock
import pytest
from io import StringIO
from pyramid.httpexceptions import HTTPOk
from kinto.core.decorators import cache_forever


@cache_forever
def demo1(request):
    request.mock()
    return "demo1"


@cache_forever
def demo2(request, name):
    return "demo2: {}".format(name)


@cache_forever
def demo3(request):
    return request.response


def test_cache_forever_decorator_call_the_decorated_function_once():
    request = mock.MagicMock()
    demo1(request)
    demo1(request)
    demo1(request)
    assert request.mock.call_count == 1


def test_cache_forever_doesnt_care_about_arguments():
    request1 = mock.MagicMock()
    request1.response = StringIO()

    request2 = mock.MagicMock()
    request2.response = StringIO()

    response1 = demo2(request1, 'Henri').getvalue()
    response2 = demo2(request2, 'Paul').getvalue()
    assert response1 == response2 == 'demo2: Henri'


def test_each_function_is_cached_separately_for_the_life_of_the_process():
    request1 = mock.MagicMock()
    request1.response = StringIO()
    response1 = demo1(request1).getvalue()

    request2 = mock.MagicMock()
    request2.response = StringIO()
    response2 = demo2(request2).getvalue()

    assert response1 != response2
    assert response1 == 'demo1'
    assert response2 == 'demo2: Henri'


def test_should_not_cache_responses():
    request = mock.MagicMock()
    request.response = HTTPOk()

    with pytest.raises(ValueError):
        demo3(request)



import mock

from kinto.core.utils import sqlalchemy
from kinto.core.storage import generators, memory, postgresql, exceptions, StorageBase
from kinto.core.testing import (unittest, skip_if_no_postgresql)
from kinto.core.storage.testing import StorageTest


class GeneratorTest(unittest.TestCase):
    def test_generic_has_mandatory_override(self):
        self.assertRaises(NotImplementedError, generators.Generator)

    def test_id_generator_must_respect_storage_backends(self):
        class Dumb(generators.Generator):
            def __call__(self):
                return '*' * 80

        self.assertRaises(ValueError, Dumb)

    def test_default_generator_allow_underscores_dash_alphabet(self):
        class Dumb(generators.Generator):
            def __call__(self):
                return '1234'

        generator = Dumb()
        self.assertTrue(generator.match('1_2_3-abc'))
        self.assertTrue(generator.match('abc_123'))
        self.assertFalse(generator.match('-1_2_3-abc'))
        self.assertFalse(generator.match('_1_2_3-abc'))

    def test_uuid_generator_pattern_allows_uuid_only(self):
        invalid_uuid = 'XXX-00000000-0000-5000-a000-000000000000'
        generator = generators.UUID4()
        self.assertFalse(generator.match(invalid_uuid))

    def test_uuid_generator_pattern_is_not_restricted_to_uuid4(self):
        generator = generators.UUID4()
        valid_uuid = 'fd800e8d-e8e9-3cac-f502-816cbed9bb6c'
        self.assertTrue(generator.match(valid_uuid))
        invalid_uuid4 = '00000000-0000-5000-a000-000000000000'
        self.assertTrue(generator.match(invalid_uuid4))
        invalid_uuid4 = '00000000-0000-4000-e000-000000000000'
        self.assertTrue(generator.match(invalid_uuid4))


class StorageBaseTest(unittest.TestCase):
    def setUp(self):
        self.storage = StorageBase()

    def test_mandatory_overrides(self):
        calls = [
            (self.storage.initialize_schema,),
            (self.storage.flush,),
            (self.storage.collection_timestamp, '', ''),
            (self.storage.create, '', '', {}),
            (self.storage.get, '', '', ''),
            (self.storage.update, '', '', '', {}),
            (self.storage.delete, '', '', ''),
            (self.storage.delete_all, '', ''),
            (self.storage.purge_deleted, '', ''),
            (self.storage.get_all, '', ''),
        ]
        for call in calls:
            self.assertRaises(NotImplementedError, *call)

    def test_backend_error_message_provides_given_message_if_defined(self):
        error = exceptions.BackendError(message="Connection Error")
        self.assertEqual(str(error), "Connection Error")

    def test_backenderror_message_default_to_original_exception_message(self):
        error = exceptions.BackendError(ValueError("Pool Error"))
        self.assertEqual(str(error), "ValueError: Pool Error")


class MemoryStorageTest(StorageTest, unittest.TestCase):
    backend = memory

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.storage,
            '_bump_timestamp',
            side_effect=exceptions.BackendError("Segmentation fault."))

    def test_backend_error_provides_original_exception(self):
        pass

    def test_raises_backend_error_if_error_occurs_on_client(self):
        pass

    def test_backend_error_is_raised_anywhere(self):
        pass

    def test_backenderror_message_default_to_original_exception_message(self):
        pass

    def test_ping_logs_error_if_unavailable(self):
        pass


@skip_if_no_postgresql
class PostgreSQLStorageTest(StorageTest, unittest.TestCase):
    backend = postgresql
    settings = {
        'storage_max_fetch_size': 10000,
        'storage_backend': 'kinto.core.storage.postgresql',
        'storage_poolclass': 'sqlalchemy.pool.StaticPool',
        'storage_url': 'postgres://postgres:postgres@localhost:5432/testdb',
    }

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.storage.client,
            'session_factory',
            side_effect=sqlalchemy.exc.SQLAlchemyError)

    def test_number_of_fetched_records_can_be_limited_in_settings(self):
        for i in range(4):
            self.create_record({'phone': 'tel-{}'.format(i)})

        results, count = self.storage.get_all(**self.storage_kw)
        self.assertEqual(len(results), 4)

        settings = {**self.settings, 'storage_max_fetch_size': 2}
        config = self._get_config(settings=settings)
        limited = self.backend.load_from_config(config)

        results, count = limited.get_all(**self.storage_kw)
        self.assertEqual(len(results), 2)

    def test_connection_is_rolledback_if_error_occurs(self):
        with self.storage.client.connect() as conn:
            query = "DELETE FROM metadata WHERE name = 'roll';"
            conn.execute(query)

        try:
            with self.storage.client.connect() as conn:
                query = "INSERT INTO metadata VALUES ('roll', 'back');"
                conn.execute(query)
                conn.commit()

                query = "INSERT INTO metadata VALUES ('roll', 'rock');"
                conn.execute(query)

                raise sqlalchemy.exc.TimeoutError()
        except exceptions.BackendError:
            pass

        with self.storage.client.connect() as conn:
            query = "SELECT COUNT(*) FROM metadata WHERE name = 'roll';"
            result = conn.execute(query)
            self.assertEqual(result.fetchone()[0], 1)

    def test_pool_object_is_shared_among_backend_instances(self):
        config = self._get_config()
        storage1 = self.backend.load_from_config(config)
        storage2 = self.backend.load_from_config(config)
        self.assertEqual(id(storage1.client),
                         id(storage2.client))

    def test_warns_if_configured_pool_size_differs_for_same_backend_type(self):
        self.backend.load_from_config(self._get_config())
        settings = {**self.settings, 'storage_pool_size': 1}
        msg = ('Reuse existing PostgreSQL connection. Parameters storage_* '
               'will be ignored.')
        with mock.patch('kinto.core.storage.postgresql.client.'
                        'warnings.warn') as mocked:
            self.backend.load_from_config(self._get_config(settings=settings))
            mocked.assert_any_call(msg)

    def test_integrity_error_rollsback_transaction(self):
        client = postgresql.create_from_config(self._get_config(),
                                               prefix='storage_',
                                               with_transaction=False)
        with self.assertRaises(exceptions.IntegrityError):
            with client.connect() as conn:
                # Make some change in metadata.
                conn.execute("INSERT INTO metadata VALUES ('roll', 'rock');")
                # Go into a failing integrity constraint.
                query = "INSERT INTO timestamps VALUES ('a', 'b', NOW());"
                conn.execute(query)
                conn.execute(query)
                conn.commit()
                conn.close()

        # Check that change in metadata was rolledback.
        with client.connect() as conn:
            result = conn.execute("SELECT FROM metadata WHERE name = 'roll';")
        self.assertEqual(result.rowcount, 0)



import pkg_resources
import logging

import kinto.core
from pyramid.config import Configurator
from pyramid.settings import asbool
from pyramid.security import Authenticated, Everyone

from kinto.authorization import RouteFactory


# Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution(__package__).version

# Implemented HTTP API Version
HTTP_API_VERSION = '1.15'

# Main kinto logger
logger = logging.getLogger(__name__)


DEFAULT_SETTINGS = {
    'flush_endpoint_enabled': False,
    'retry_after_seconds': 3,
    'cache_backend': 'kinto.core.cache.memory',
    'permission_backend': 'kinto.core.permission.memory',
    'storage_backend': 'kinto.core.storage.memory',
    'project_docs': 'https://kinto.readthedocs.io/',
    'bucket_create_principals': Authenticated,
    'permissions_read_principals': Everyone,
    'multiauth.authorization_policy': (
        'kinto.authorization.AuthorizationPolicy'),
    'experimental_collection_schema_validation': False,
    'experimental_permissions_endpoint': False,
    'http_api_version': HTTP_API_VERSION,
    'bucket_id_generator': 'kinto.views.NameGenerator',
    'collection_id_generator': 'kinto.views.NameGenerator',
    'group_id_generator': 'kinto.views.NameGenerator',
    'record_id_generator': 'kinto.views.RelaxedUUID'
}


def main(global_config, config=None, **settings):
    if not config:
        config = Configurator(settings=settings, root_factory=RouteFactory)

    # Force project name, since it determines settings prefix.
    config.add_settings({'kinto.project_name': 'kinto'})

    kinto.core.initialize(config,
                          version=__version__,
                          default_settings=DEFAULT_SETTINGS)

    settings = config.get_settings()

    # Expose capability
    schema_enabled = asbool(
        settings['experimental_collection_schema_validation']
    )
    if schema_enabled:
        config.add_api_capability(
            "schema",
            description="Validates collection records with JSON schemas.",
            url="https://kinto.readthedocs.io/en/latest/api/1.x/"
                "collections.html#collection-json-schema")

    # Scan Kinto views.
    kwargs = {}

    flush_enabled = asbool(settings['flush_endpoint_enabled'])
    if flush_enabled:
        config.add_api_capability(
            "flush_endpoint",
            description="The __flush__ endpoint can be used to remove all "
                        "data from all backends.",
            url="https://kinto.readthedocs.io/en/latest/configuration/"
                "settings.html#activating-the-flush-endpoint")
    else:
        kwargs['ignore'] = ['kinto.views.flush']

    # Permissions endpoint enabled if permission backend is setup.
    permissions_endpoint_enabled = (
        asbool(settings['experimental_permissions_endpoint']) and
        hasattr(config.registry, 'permission'))
    if permissions_endpoint_enabled:
        config.add_api_capability(
            "permissions_endpoint",
            description="The permissions endpoint can be used to list all "
                        "user objects permissions.",
            url="https://kinto.readthedocs.io/en/latest/configuration/"
                "settings.html#activating-the-permissions-endpoint")
    else:
        kwargs.setdefault('ignore', []).append('kinto.views.permissions')

    config.scan("kinto.views", **kwargs)

    app = config.make_wsgi_app()

    # Install middleware (no-op if disabled)
    return kinto.core.install_middlewares(app, settings)



from bravado_core.response import validate_response

from .support import SwaggerTest, MINIMALIST_COLLECTION


class SwaggerCollectionResponsesTest(SwaggerTest):

    def test_get_collection_200(self):
        response = self.app.get('/buckets/b1/collections/c1',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].get_collection
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_post_collection_200(self):
        response = self.app.post_json('/buckets/b1/collections', self.collection,
                                      headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].create_collection
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_post_collection_201(self):
        response = self.app.post_json('/buckets/b1/collections', MINIMALIST_COLLECTION,
                                      headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].create_collection
        schema = self.spec.deref(op.op_spec['responses']['201'])
        validate_response(schema, op, response)

    def test_put_collection_200(self):
        response = self.app.put_json('/buckets/b1/collections/c1', MINIMALIST_COLLECTION,
                                     headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].update_collection
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_put_collection_201(self):
        response = self.app.put_json('/buckets/b1/collections/c2', MINIMALIST_COLLECTION,
                                     headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].update_collection
        schema = self.spec.deref(op.op_spec['responses']['201'])
        validate_response(schema, op, response)

    def test_delete_collection_200(self):
        response = self.app.delete('/buckets/b1/collections/c1',
                                   headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].delete_collection
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_get_collections_200(self):
        response = self.app.get('/buckets/b1/collections',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].get_collections
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_delete_collections_200(self):
        response = self.app.delete('/buckets/b1/collections',
                                   headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Collections'].delete_collections
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)



import contextlib
import warnings
from collections import defaultdict

from kinto.core import logger
from kinto.core.storage import exceptions
from kinto.core.utils import sqlalchemy
import transaction as zope_transaction


class PostgreSQLClient:
    def __init__(self, session_factory, commit_manually, invalidate):
        self.session_factory = session_factory
        self.commit_manually = commit_manually
        self.invalidate = invalidate

    @contextlib.contextmanager
    def connect(self, readonly=False, force_commit=False):
        """
        Pulls a connection from the pool when context is entered and
        returns it when context is exited.

        A COMMIT is performed on the current transaction if everything went
        well. Otherwise transaction is ROLLBACK, and everything cleaned up.
        """
        commit_manually = self.commit_manually and not readonly
        session = None
        try:
            # Pull connection from pool.
            session = self.session_factory()
            # Start context
            yield session
            if not readonly and not self.commit_manually:
                # Mark session as dirty.
                self.invalidate(session)
            # Success
            if commit_manually:
                session.commit()
            elif force_commit:
                # Commit like would do a succesful request.
                zope_transaction.commit()

        except sqlalchemy.exc.IntegrityError as e:
            logger.error(e)
            if commit_manually:  # pragma: no branch
                session.rollback()
            raise exceptions.IntegrityError(original=e) from e
        except sqlalchemy.exc.SQLAlchemyError as e:
            logger.error(e)
            if session and commit_manually:
                session.rollback()
            raise exceptions.BackendError(original=e) from e
        finally:
            if session and self.commit_manually:
                # Give back to pool if commit done manually.
                session.close()


# Reuse existing client if same URL.
_CLIENTS = defaultdict(dict)


def create_from_config(config, prefix='', with_transaction=True):
    """Create a PostgreSQLClient client using settings in the provided config.
    """
    if sqlalchemy is None:
        message = ("PostgreSQL SQLAlchemy dependency missing. "
                   "Refer to installation section in documentation.")
        raise ImportWarning(message)

    from zope.sqlalchemy import ZopeTransactionExtension, invalidate
    from sqlalchemy.orm import sessionmaker, scoped_session

    settings = {**config.get_settings()}
    # Custom Kinto settings, unsupported by SQLAlchemy.
    settings.pop(prefix + 'backend', None)
    settings.pop(prefix + 'max_fetch_size', None)
    settings.pop(prefix + 'max_size_bytes', None)
    settings.pop(prefix + 'prefix', None)
    transaction_per_request = with_transaction and settings.pop('transaction_per_request', False)

    url = settings[prefix + 'url']
    existing_client = _CLIENTS[transaction_per_request].get(url)
    if existing_client:
        msg = ("Reuse existing PostgreSQL connection. "
               "Parameters {}* will be ignored.".format(prefix))
        warnings.warn(msg)
        return existing_client

    # Initialize SQLAlchemy engine from settings.
    poolclass_key = prefix + 'poolclass'
    settings.setdefault(poolclass_key, ('kinto.core.storage.postgresql.'
                                        'pool.QueuePoolWithMaxBacklog'))
    settings[poolclass_key] = config.maybe_dotted(settings[poolclass_key])
    engine = sqlalchemy.engine_from_config(settings, prefix=prefix, url=url)

    # Initialize thread-safe session factory.
    options = {}
    if transaction_per_request:
        # Plug with Pyramid transaction manager
        options['extension'] = ZopeTransactionExtension()
    session_factory = scoped_session(sessionmaker(bind=engine, **options))

    # Store one client per URI.
    commit_manually = (not transaction_per_request)
    client = PostgreSQLClient(session_factory, commit_manually, invalidate)
    _CLIENTS[transaction_per_request][url] = client
    return client



import unittest

from kinto.core.testing import get_user_headers

from .support import (BaseWebTest,
                      MINIMALIST_BUCKET, MINIMALIST_COLLECTION,
                      MINIMALIST_GROUP, MINIMALIST_RECORD)


class PermissionsTest(BaseWebTest, unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alice_headers = {**self.headers, **get_user_headers('alice')}
        self.bob_headers = {**self.headers, **get_user_headers('bob')}

        self.alice_principal = ('basicauth:d5b0026601f1b251974e09548d44155e16'
                                '812e3c64ff7ae053fe3542e2ca1570')
        self.bob_principal = ('basicauth:c031ced27503f788b102ca54269a062ec737'
                              '94bb075154c74a0d4311e74ca8b6')


class BucketPermissionsTest(PermissionsTest):

    def setUp(self):
        bucket = {**MINIMALIST_BUCKET, 'permissions': {'read': [self.alice_principal]}}
        self.app.put_json('/buckets/sodas',
                          bucket,
                          headers=self.headers)

    def test_creation_is_allowed_to_authenticated_by_default(self):
        self.app.put_json('/buckets/beer',
                          MINIMALIST_BUCKET,
                          headers=self.headers)

    def test_current_user_receives_write_permission_on_creation(self):
        resp = self.app.put_json('/buckets/beer',
                                 MINIMALIST_BUCKET,
                                 headers=self.headers)
        permissions = resp.json['permissions']
        self.assertIn(self.principal, permissions['write'])

    def test_can_read_if_allowed(self):
        self.app.get('/buckets/sodas',
                     headers=self.alice_headers)

    def test_cannot_write_if_not_allowed(self):
        self.app.put_json('/buckets/sodas',
                          MINIMALIST_BUCKET,
                          headers=self.alice_headers,
                          status=403)

    def test_permissions_are_not_returned_if_can_only_read(self):
        resp = self.app.get('/buckets/sodas', headers=self.alice_headers)
        self.assertEqual(resp.json['permissions'], {})

    def test_permissions_are_returned_if_can_write(self):
        resp = self.app.get('/buckets/sodas', headers=self.headers)
        self.assertIn('write', resp.json['permissions'])


class CollectionPermissionsTest(PermissionsTest):

    def setUp(self):
        bucket = {**MINIMALIST_BUCKET, 'permissions': {
            'read': [self.alice_principal],
            'write': [self.bob_principal]
        }}
        self.app.put_json('/buckets/beer',
                          bucket,
                          headers=self.headers)
        self.app.put_json('/buckets/beer/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)

    def test_passing_unicode_on_parent_id_is_supported(self):
        self.app.get('/buckets/block%C2%93%C2%96sts/collections/barley',
                     headers=self.alice_headers,
                     status=403)

    def test_read_is_allowed_if_read_on_bucket(self):
        self.app.get('/buckets/beer/collections/barley',
                     headers=self.alice_headers)

    def test_read_is_allowed_if_write_on_bucket(self):
        self.app.get('/buckets/beer/collections/barley',
                     headers=self.bob_headers)

    def test_cannot_read_if_not_allowed(self):
        headers = {**self.headers, **get_user_headers('jean-louis')}
        self.app.get('/buckets/beer/collections/barley',
                     headers=headers,
                     status=403)

    def test_cannot_write_if_not_allowed(self):
        self.app.put_json('/buckets/beer/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.alice_headers,
                          status=403)

    def test_permission_backend_prevent_sql_injections(self):
        self.app.get("/buckets/beer'", headers=self.headers, status=403)
        self.app.get("/buckets/beer'/collections/barley", headers=self.headers, status=403)
        self.app.get("/buckets/beer'/groups/barley", headers=self.headers, status=403)

        self.app.get("/buckets/beer/collections/barley'", headers=self.headers, status=400)
        # XXX: We should validate the collection ID on the records collection endpoint. #1077
        self.app.get("/buckets/beer/collections/barley'/records", headers=self.headers, status=404)

        self.app.get("/buckets/beer/groups/barley'", headers=self.headers, status=400)


class GroupPermissionsTest(PermissionsTest):

    def setUp(self):
        bucket = {**MINIMALIST_BUCKET, 'permissions': {
            'read': [self.alice_principal],
            'write': [self.bob_principal]
        }}
        self.app.put_json('/buckets/beer',
                          bucket,
                          headers=self.headers)

        self.app.put_json('/buckets/beer/groups/moderators',
                          MINIMALIST_GROUP,
                          headers=self.headers)

    def test_creation_is_allowed_if_write_on_bucket(self):
        self.app.post_json('/buckets/beer/groups',
                           MINIMALIST_GROUP,
                           headers=self.headers)

    def test_read_is_allowed_if_read_on_bucket(self):
        self.app.get('/buckets/beer/groups/moderators',
                     headers=self.alice_headers)

    def test_read_is_allowed_if_write_on_bucket(self):
        self.app.get('/buckets/beer/groups/moderators',
                     headers=self.bob_headers)

    def test_cannot_read_if_not_allowed(self):
        headers = {**self.headers, **get_user_headers('jean-louis')}
        self.app.get('/buckets/beer/groups/moderators',
                     headers=headers,
                     status=403)

    def test_cannot_write_if_not_allowed(self):
        self.app.put_json('/buckets/beer/groups/moderators',
                          MINIMALIST_GROUP,
                          headers=self.alice_headers,
                          status=403)

    def test_creation_is_forbidden_is_no_write_on_bucket(self):
        self.app.post_json('/buckets/beer/groups',
                           MINIMALIST_GROUP,
                           headers=self.alice_headers,
                           status=403)


class RecordPermissionsTest(PermissionsTest):

    def setUp(self):
        bucket = {**MINIMALIST_BUCKET, 'permissions': {'write': [self.alice_principal]}}
        self.app.put_json('/buckets/beer',
                          bucket,
                          headers=self.headers)

        collection = {**MINIMALIST_COLLECTION, 'permissions': {'write': [self.bob_principal]}}
        self.app.put_json('/buckets/beer/collections/barley',
                          collection,
                          headers=self.headers)

    def test_creation_is_allowed_if_write_on_bucket(self):
        self.app.post_json('/buckets/beer/collections/barley/records',
                           MINIMALIST_RECORD,
                           headers=self.alice_headers)

    def test_creation_is_allowed_if_write_on_collection(self):
        self.app.post_json('/buckets/beer/collections/barley/records',
                           MINIMALIST_RECORD,
                           headers=self.bob_headers)

    def test_creation_is_forbidden_is_no_write_on_bucket_nor_collection(self):
        headers = {**self.headers, **get_user_headers('jean-louis')}
        self.app.post_json('/buckets/beer/collections/barley/records',
                           MINIMALIST_RECORD,
                           headers=headers,
                           status=403)

    def test_record_permissions_are_modified_by_patch(self):
        collection_url = '/buckets/beer/collections/barley/records'
        resp = self.app.post_json(collection_url,
                                  MINIMALIST_RECORD,
                                  headers=self.headers)
        record = resp.json['data']
        resp = self.app.patch_json('{}/{}'.format(collection_url, record['id']),
                                   {'permissions': {'read': ['fxa:user']}},
                                   headers=self.headers)
        self.assertIn('fxa:user', resp.json['permissions']['read'])


class ChildrenCreationTest(PermissionsTest):
    def setUp(self):
        self.app.put_json('/buckets/create',
                          {'permissions': {'group:create': ['system.Authenticated']}},
                          headers=self.alice_headers)
        self.app.put_json('/buckets/write',
                          {'permissions': {'write': ['system.Authenticated']}},
                          headers=self.alice_headers)
        self.app.put_json('/buckets/read',
                          {'permissions': {'read': ['system.Authenticated']}},
                          headers=self.alice_headers)
        for parent in ('create', 'write', 'read'):
            self.app.put_json('/buckets/{}/groups/child'.format(parent),
                              MINIMALIST_GROUP,
                              headers=self.alice_headers)
        self.bob_headers_safe_creation = dict({'If-None-Match': '*'},
                                              **self.bob_headers)

    def test_cannot_read_others_objects_if_only_allowed_to_create(self):
        self.app.get('/buckets/create/groups/child', headers=self.bob_headers, status=403)

    def test_safe_creation_with_put_returns_412_if_allowed_to_create(self):
        self.app.put_json('/buckets/create/groups/child',
                          MINIMALIST_GROUP,
                          headers=self.bob_headers_safe_creation, status=412)

    def test_safe_creation_with_post_returns_412_if_allowed_to_create(self):
        self.app.post_json('/buckets/create/groups',
                           {'data': {'id': 'child', 'members': []}},
                           headers=self.bob_headers_safe_creation, status=412)

    def test_safe_creation_with_put_returns_412_if_allowed_to_write(self):
        self.app.put_json('/buckets/write/groups/child',
                          MINIMALIST_GROUP,
                          headers=self.bob_headers_safe_creation, status=412)

    def test_safe_creation_with_post_returns_412_if_allowed_to_write(self):
        self.app.post_json('/buckets/write/groups',
                           {'data': {'id': 'child', 'members': []}},
                           headers=self.bob_headers_safe_creation, status=412)

    def test_safe_creation_with_put_returns_403_if_only_allowed_to_read(self):
        self.app.put_json('/buckets/read/groups/child',
                          MINIMALIST_GROUP,
                          headers=self.bob_headers_safe_creation, status=403)

    def test_safe_creation_with_post_returns_403_if_only_allowed_to_read(self):
        self.app.post_json('/buckets/read/groups',
                           {'data': {'id': 'child', 'members': []}},
                           headers=self.bob_headers_safe_creation, status=403)

    def test_delete_returns_404_on_unknown_if_only_allowed_to_read(self):
        self.app.delete('/buckets/read/groups/g1',
                        headers=self.bob_headers,
                        status=404)

    def test_patch_returns_404_on_unknown_if_only_allowed_to_read(self):
        self.app.patch_json('/buckets/read/groups/g1',
                            {'data': {'members': []}},
                            headers=self.bob_headers,
                            status=404)


class ParentMetadataTest(PermissionsTest):
    def setUp(self):
        self.app.put_json('/buckets/beer',
                          {'permissions': {'collection:create': [self.bob_principal]}},
                          headers=self.headers)

        self.app.put_json('/buckets/beer/collections/wheat', headers=self.headers)
        self.app.put_json('/buckets/beer/collections/root', headers=self.headers)

        self.app.put_json('/buckets/beer/collections/barley',
                          {'permissions': {'record:create': [self.alice_principal]}},
                          headers=self.bob_headers)

    def test_parent_metadata_can_be_read_if_allowed_to_create_child(self):
        self.app.get('/buckets/beer', headers=self.bob_headers)
        self.app.get('/buckets/beer/collections/barley', headers=self.alice_headers)

    def test_parent_metadata_cannot_be_read_if_not_allowed_to_create_child(self):
        self.app.get('/buckets/beer',
                     headers=get_user_headers('jean:paul'),
                     status=403)
        self.app.get('/buckets/beer/collections/barley',
                     headers=get_user_headers('mahmud:hatim'),
                     status=403)

    def test_list_can_be_obtained_if_allowed_to_create(self):
        resp = self.app.get('/buckets/beer/collections', headers=self.bob_headers)
        self.assertEqual(len(resp.json['data']), 1)
        self.assertEqual(resp.json['data'][0]['id'], 'barley')

        resp = self.app.get('/buckets/beer/collections/barley/records', headers=self.alice_headers)
        self.assertEqual(resp.json['data'], [])

    def test_list_is_denied_if_not_allowed_to_create(self):
        self.app.get('/buckets/beer/collections',
                     headers=get_user_headers('jean:paul'),
                     status=403)
        self.app.get('/buckets/beer/collections/barley/records',
                     headers=get_user_headers('mahmud:hatim'),
                     status=403)



import unittest
import os
import pytest

import colander
import mock
from kinto.core import includeme
from kinto.core import DEFAULT_SETTINGS
from pyramid import httpexceptions
from pyramid import request as pyramid_request
from pyramid import testing

from kinto.core.utils import (
    native_value, strip_whitespace, random_bytes_hex, read_env, hmac_digest,
    current_service, follow_subrequest, build_request, dict_subset, dict_merge,
    parse_resource, prefixed_principals, recursive_update_dict,
    find_nested_value
)
from kinto.core.testing import DummyRequest


def build_real_request(wsgi_environ):
    """Build a Pyramid request, as if it was instantiated by Pyramid.
    """
    config = testing.setUp(settings=DEFAULT_SETTINGS)
    includeme(config)
    request = pyramid_request.Request(wsgi_environ)
    request.registry = config.registry
    return request


class NativeValueTest(unittest.TestCase):
    def test_simple_string(self):
        self.assertEqual(native_value('value'), 'value')

    def test_integer(self):
        self.assertEqual(native_value('7'), 7)

    def test_zero_and_one_coerce_to_integers(self):
        self.assertEqual(native_value('1'), 1)
        self.assertEqual(native_value('0'), 0)

    def test_float(self):
        self.assertEqual(native_value('3.14'), 3.14)

    def test_true_values(self):
        true_strings = ['True', 'on', 'true', 'yes']
        true_values = [native_value(s) for s in true_strings]
        self.assertTrue(all(true_values))

    def test_false_values(self):
        false_strings = ['False', 'off', 'false', 'no']
        false_values = [native_value(s) for s in false_strings]
        self.assertFalse(any(false_values))

    def test_non_string_values(self):
        self.assertEqual(native_value(7), 7)
        self.assertEqual(native_value(True), True)

    def test_bad_string_values(self):
        self.assertEqual(native_value("\u0000"), "\x00")


class StripWhitespaceTest(unittest.TestCase):
    def test_removes_all_kinds_of_spaces(self):
        value = " \t teaser \n \r"
        self.assertEqual(strip_whitespace(value), 'teaser')

    def test_does_remove_middle_spaces(self):
        self.assertEqual(strip_whitespace('a b c'), 'a b c')

    def test_idempotent_for_null_values(self):
        self.assertEqual(strip_whitespace(colander.null), colander.null)


class CryptographicRandomBytesTest(unittest.TestCase):
    def test_return_hex_string(self):
        value = random_bytes_hex(16)
        try:
            int(value, 16)
        except ValueError:
            self.fail("{} is not an hexadecimal value.".format(value))

    def test_return_right_length_string(self):
        for x in range(2, 4):
            value = random_bytes_hex(x)
            self.assertEqual(len(value), x * 2)

    def test_return_text_string(self):
        value = random_bytes_hex(16)
        self.assertIsInstance(value, str)


class HmacDigestTest(unittest.TestCase):
    def test_supports_secret_as_text(self):
        value = hmac_digest("blah", "input data")
        self.assertTrue(value.startswith("d4f5c51db246c7faeb42240545b47274b6"))

    def test_supports_secret_as_bytes(self):
        value = hmac_digest(b"blah", "input data")
        self.assertTrue(value.startswith("d4f5c51db246c7faeb42240545b47274b6"))


class ReadEnvironmentTest(unittest.TestCase):
    def test_return_passed_value_if_not_defined_in_env(self):
        self.assertEqual(read_env('missing', 12), 12)

    def test_return_env_value_if_defined_in_env(self):
        os.environ.setdefault('KINTO_CONF', 'abc')
        self.assertEqual(read_env('KINTO_CONF', 12), 'abc')

    def test_return_env_name_as_uppercase(self):
        os.environ.setdefault('KINTO_NAME', 'abc')
        self.assertEqual(read_env('kinto.name', 12), 'abc')

    def test_return_env_value_is_coerced_to_python(self):
        os.environ.setdefault('KINTO_CONF_NAME', '3.14')
        self.assertEqual(read_env('kinto-conf.name', 12), 3.14)


class CurrentServiceTest(unittest.TestCase):

    def test_current_service_returns_the_service_for_existing_patterns(self):
        request = DummyRequest()
        request.matched_route.pattern = '/buckets'
        request.registry.cornice_services = {'/buckets': mock.sentinel.service}

        self.assertEqual(current_service(request), mock.sentinel.service)

    def test_current_service_returns_none_for_unexisting_patterns(self):
        request = DummyRequest()
        request.matched_route.pattern = '/unexisting'
        request.registry.cornice_services = {}

        self.assertEqual(current_service(request), None)


class PrefixedPrincipalsTest(unittest.TestCase):

    def test_removes_unprefixed_from_principals(self):
        request = DummyRequest()
        request.effective_principals = ['foo', 'system.Authenticated']
        request.prefixed_userid = 'basic:foo'
        self.assertEqual(prefixed_principals(request),
                         ['system.Authenticated', 'basic:foo'])

    def test_works_if_userid_is_not_in_principals(self):
        request = DummyRequest()
        request.effective_principals = ['basic:foo', 'system.Authenticated']
        request.prefixed_userid = 'basic:foo'
        self.assertEqual(prefixed_principals(request),
                         ['basic:foo', 'system.Authenticated'])


class BuildRequestTest(unittest.TestCase):

    def test_built_request_has_kinto_core_custom_methods(self):
        original = build_real_request({'PATH_INFO': '/foo'})
        request = build_request(original, {"path": "bar"})
        self.assertTrue(hasattr(request, 'current_service'))


class FollowSubrequestTest(unittest.TestCase):

    def test_parent_and_bound_data_are_preserved(self):
        request = DummyRequest()
        request.invoke_subrequest.side_effect = (
            httpexceptions.HTTPTemporaryRedirect, None)
        subrequest = DummyRequest()
        subrequest.parent = mock.sentinel.parent
        subrequest.bound_data = mock.sentinel.bound_data
        _, redirected = follow_subrequest(request, subrequest)
        self.assertEqual(subrequest.parent, redirected.parent)
        self.assertEqual(subrequest.bound_data, redirected.bound_data)


class DictSubsetTest(unittest.TestCase):

    def test_extract_by_keys(self):
        obtained = dict_subset(dict(a=1, b=2), ["b"])
        expected = dict(b=2)
        self.assertEqual(obtained, expected)

    def test_is_noop_if_no_keys(self):
        obtained = dict_subset(dict(a=1, b=2), [])
        expected = dict()
        self.assertEqual(obtained, expected)

    def test_ignores_unknown_keys(self):
        obtained = dict_subset(dict(a=1, b=2), ["a", "a.b", "d.b", "c"])
        expected = dict(a=1)
        self.assertEqual(obtained, expected)

    def test_ignores_duplicated_keys(self):
        obtained = dict_subset(dict(a=1, b=2), ["a", "a"])
        expected = dict(a=1)
        self.assertEqual(obtained, expected)

    def test_can_filter_subobjects(self):
        obtained = dict_subset(dict(a=1, b=dict(c=2, d=3)), ["a", "b.c"])
        expected = dict(a=1, b=dict(c=2))
        self.assertEqual(obtained, expected)

    def test_can_filter_subobjects_keys(self):
        input = dict(a=1, b=dict(c=2, d=3, e=4))
        obtained = dict_subset(input, ["a", "b.d", "b.e"])
        expected = dict(a=1, b=dict(d=3, e=4))
        self.assertEqual(obtained, expected)

    def test_can_filter_subobjects_recursively(self):
        input = dict(b=dict(c=2, d=dict(e=4, f=5, g=6)))
        obtained = dict_subset(input, ["b.d.e", "b.d.f"])
        expected = dict(b=dict(d=dict(e=4, f=5)))
        self.assertEqual(obtained, expected)

    def test_ignores_if_subobject_is_not_dict(self):
        input = dict(a=1, b=dict(c=2, d=3))
        obtained = dict_subset(input, ["a", "b.c.d", "b.d"])
        expected = dict(a=1, b=dict(c=2, d=3))
        self.assertEqual(obtained, expected)


class DictMergeTest(unittest.TestCase):

    def test_merge(self):
        obtained = dict_merge(dict(a=1, b=dict(c=2)), dict(b=dict(d=4)))
        expected = dict(a=1, b=dict(c=2, d=4))
        self.assertEqual(obtained, expected)


class FindNestedValueTest(unittest.TestCase):

    def test_find_flat_value(self):
        record = {"a": 42}
        obtained = find_nested_value(record, "a")
        self.assertEqual(obtained, 42)

    def test_find_nested_value(self):
        record = {"a": {"b": 42}}
        obtained = find_nested_value(record, "a.b")
        self.assertEqual(obtained, 42)

    def test_find_deeply_nested_value(self):
        record = {"a": {"b": {"c": 42}}}
        obtained = find_nested_value(record, "a.b.c")
        self.assertEqual(obtained, 42)

    def test_find_dotted_path_value(self):
        record = {"a.b": 42}
        obtained = find_nested_value(record, "a.b")
        self.assertEqual(obtained, 42)

    def test_find_nested_dotted_path_value(self):
        record = {"a": {"b.c": 42}}
        obtained = find_nested_value(record, "a.b.c")
        self.assertEqual(obtained, 42)

        record = {"a.b": {"c.d": 42}}
        obtained = find_nested_value(record, "a.b.c.d")
        self.assertEqual(obtained, 42)

        record = {"a": {"b": {"a": {"b": 42}}}}
        obtained = find_nested_value(record, "a.b.a.b")
        self.assertEqual(obtained, 42)

    def test_find_disambiguated_dotted_path_values(self):
        # XXX: To be honest, this is a really scary use case. Probably a
        # limitation of the dotted path notation we may want to document.
        # At least this test acts as documentation for now.
        record = {"a": {"b": 0}, "a.b": 42}
        obtained = find_nested_value(record, "a.b")
        self.assertEqual(obtained, 42)

    def test_unmatched_path_returns_none(self):
        record = {"a": 42}
        self.assertIsNone(find_nested_value(record, "x"))
        self.assertIsNone(find_nested_value(record, "a.b"))
        self.assertIsNone(find_nested_value(record, "x.a"))

    def test_fallback_default_value(self):
        record = {"a": {"c": 42}}
        self.assertEqual(find_nested_value(record, "x", 1337), 1337)
        self.assertEqual(find_nested_value(record, "a.b", 1337), 1337)
        self.assertEqual(find_nested_value(record, "x.a", 1337), 1337)
        self.assertEqual(find_nested_value(record, "a.c.d", 1337), 1337)


class RecursiveUpdateDictTest(unittest.TestCase):

    def test_merge(self):
        a = {}
        recursive_update_dict(a, {'b': {'c': 1}, 'd': 2})
        self.assertEqual(a['b']['c'], 1)
        self.assertEqual(a['d'], 2)

    def test_merge_non_dict(self):
        a = {}
        recursive_update_dict(a, 1)
        self.assertEqual(a, {})


class ParseResourceTest(unittest.TestCase):

    expected = {
        'bucket': 'bid',
        'collection': 'cid'
    }
    error_msg = "Resources should be defined as "
    "'/buckets/<bid>/collections/<cid>' or '<bid>/<cid>'. "
    "with valid collection and bucket ids."

    def _assert_success(self, input):
        parts = parse_resource(input)
        self.assertEqual(self.expected, parts)

    def _assert_error(self, input_arr):
        for input in input_arr:
            with pytest.raises(ValueError) as excinfo:
                parse_resource(input)
            self.assertEquals(str(excinfo.value), self.error_msg)

    def test_malformed_url_raises_an_exception(self):
        input_arr = ['foo', '/', '/bar', 'baz/', '/fo/o', '/buckets/sbid/scid']
        self._assert_error(input_arr)

    def test_returned_resources_match_the_expected_format(self):
        input = "/buckets/bid/collections/cid"
        self._assert_success(input)

    def test_returned_resources_match_the_legacy_format(self):
        input = "bid/cid"
        self._assert_success(input)

    def test_resources_must_be_valid_names(self):
        input_arr = ['/buckets/bi+d1/collections/cid', '/buckets/bid1/collections/dci,d']
        self._assert_error(input_arr)



import json
import unittest

from kinto.core.errors import send_alert
from kinto.core.testing import DummyRequest


class SendAlertTest(unittest.TestCase):

    def verify_alert_header(self, request, expected):
        self.assertIn('Alert', request.response.headers)
        alert = request.response.headers['Alert']
        self.assertDictEqual(json.loads(alert),
                             expected)

    def test_send_alert_default_to_project_url(self):
        request = DummyRequest()
        request.registry.settings['project_docs'] = 'docs_url'
        send_alert(request, 'Message')
        self.verify_alert_header(request, {
            'code': 'soft-eol',
            'message': 'Message',
            'url': 'docs_url'
        })

    def test_send_alert_url_can_be_specified(self):
        request = DummyRequest()
        send_alert(request, 'Message', 'error_url')
        self.verify_alert_header(request, {
            'code': 'soft-eol',
            'message': 'Message',
            'url': 'error_url'
        })

    def test_send_alert_code_can_be_specified(self):
        request = DummyRequest()
        request.registry.settings['project_docs'] = 'docs_url'
        send_alert(request, 'Message', code='hard-eol')
        self.verify_alert_header(request, {
            'code': 'hard-eol',
            'message': 'Message',
            'url': 'docs_url'
        })



import copy

import jsonschema
from kinto.core import resource, utils
from kinto.core.errors import raise_invalid
from jsonschema import exceptions as jsonschema_exceptions
from pyramid.security import Authenticated
from pyramid.settings import asbool

from kinto.views import object_exists_or_404


_parent_path = '/buckets/{{bucket_id}}/collections/{{collection_id}}'


@resource.register(name='record',
                   collection_path=_parent_path + '/records',
                   record_path=_parent_path + '/records/{{id}}')
class Record(resource.ShareableResource):

    schema_field = 'schema'

    def __init__(self, request, **kwargs):
        # Before all, first check that the parent collection exists.
        # Check if already fetched before (in batch).
        collections = request.bound_data.setdefault('collections', {})
        collection_uri = self.get_parent_id(request)
        if collection_uri not in collections:
            # Unknown yet, fetch from storage.
            collection_parent_id = utils.instance_uri(request, 'bucket',
                                                      id=self.bucket_id)
            collection = object_exists_or_404(request,
                                              collection_id='collection',
                                              parent_id=collection_parent_id,
                                              object_id=self.collection_id)
            collections[collection_uri] = collection

        super().__init__(request, **kwargs)
        self._collection = collections[collection_uri]

    def get_parent_id(self, request):
        self.bucket_id = request.matchdict['bucket_id']
        self.collection_id = request.matchdict['collection_id']
        return utils.instance_uri(request, 'collection',
                                  bucket_id=self.bucket_id,
                                  id=self.collection_id)

    def process_record(self, new, old=None):
        """Validate records against collection schema, if any."""
        new = super().process_record(new, old)

        schema = self._collection.get('schema')
        settings = self.request.registry.settings
        schema_validation = 'experimental_collection_schema_validation'
        if not schema or not asbool(settings.get(schema_validation)):
            return new

        collection_timestamp = self._collection[self.model.modified_field]

        try:
            stripped = copy.deepcopy(new)
            stripped.pop(self.model.id_field, None)
            stripped.pop(self.model.modified_field, None)
            stripped.pop(self.model.permissions_field, None)
            stripped.pop(self.schema_field, None)
            jsonschema.validate(stripped, schema)
        except jsonschema_exceptions.ValidationError as e:
            try:
                field = e.path.pop() if e.path else e.validator_value.pop()
            except AttributeError:
                field = None
            raise_invalid(self.request, name=field, description=e.message)

        new[self.schema_field] = collection_timestamp
        return new

    def collection_get(self):
        result = super().collection_get()
        self._handle_cache_expires(self.request.response)
        return result

    def get(self):
        result = super().get()
        self._handle_cache_expires(self.request.response)
        return result

    def _handle_cache_expires(self, response):
        """If the parent collection defines a ``cache_expires`` attribute,
        then cache-control response headers are sent.

        .. note::

            Those headers are also sent if the
            ``kinto.record_cache_expires_seconds`` setting is defined.
        """
        is_anonymous = Authenticated not in self.request.effective_principals
        if not is_anonymous:
            return

        cache_expires = self._collection.get('cache_expires')
        if cache_expires is None:
            by_bucket = '{}_record_cache_expires_seconds'.format(self.bucket_id)
            by_collection = '{}_{}_record_cache_expires_seconds'.format(
                self.bucket_id, self.collection_id)
            settings = self.request.registry.settings
            cache_expires = settings.get(by_collection,
                                         settings.get(by_bucket))

        if cache_expires is not None:
            response.cache_expires(seconds=int(cache_expires))



import codecs
import mock
import os
import tempfile
import unittest
from time import strftime

from kinto import config
from kinto import __version__


class ConfigTest(unittest.TestCase):
    def test_transpose_parameters_into_template(self):
        self.maxDiff = None
        template = "kinto.tpl"
        dest = tempfile.mktemp()
        config.render_template(template, dest,
                               host='127.0.0.1',
                               secret='secret',
                               storage_backend='storage_backend',
                               cache_backend='cache_backend',
                               permission_backend='permission_backend',
                               storage_url='storage_url',
                               cache_url='cache_url',
                               permission_url='permission_url',
                               kinto_version='kinto_version',
                               config_file_timestamp='config_file_timestamp')

        with codecs.open(dest, 'r', encoding='utf-8') as d:
            destination_temp = d.read()

        sample_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                   "test_configuration/test.ini")
        with codecs.open(sample_path, 'r', encoding='utf-8') as c:
            sample = c.read()

        self.assertEqual(destination_temp, sample)

    def test_create_destination_directory(self):
        dest = os.path.join(tempfile.mkdtemp(), 'config', 'kinto.ini')

        config.render_template("kinto.tpl", dest,
                               host='127.0.0.1',
                               secret='secret',
                               storage_backend='storage_backend',
                               cache_backend='cache_backend',
                               permission_backend='permission_backend',
                               storage_url='storage_url',
                               cache_url='cache_url',
                               permission_url='permission_url',
                               kinto_version='kinto_version',
                               config_file_timestamp='config_file_timestamp')

        self.assertTrue(os.path.exists(dest))

    @mock.patch('kinto.config.render_template')
    def test_hmac_secret_is_text(self, mocked_render_template):
        config.init('kinto.ini', 'postgresql')
        args, kwargs = list(mocked_render_template.call_args)
        self.assertEquals(type(kwargs['secret']), str)

    @mock.patch('kinto.config.render_template')
    def test_init_postgresql_values(self, mocked_render_template):
        config.init('kinto.ini', 'postgresql')

        args, kwargs = list(mocked_render_template.call_args)
        self.assertEquals(args, ('kinto.tpl', 'kinto.ini'))

        postgresql_url = "postgres://postgres:postgres@localhost/postgres"
        self.assertDictEqual(kwargs, {
            'host': '127.0.0.1',
            'secret': kwargs['secret'],
            'storage_backend': 'kinto.core.storage.postgresql',
            'cache_backend': 'kinto.core.cache.postgresql',
            'permission_backend': 'kinto.core.permission.postgresql',
            'storage_url': postgresql_url,
            'cache_url':  postgresql_url,
            'permission_url': postgresql_url,
            'kinto_version': __version__,
            'config_file_timestamp': strftime('%a, %d %b %Y %H:%M:%S %z')
        })

    @mock.patch('kinto.config.render_template')
    def test_init_redis_values(self, mocked_render_template):
        config.init('kinto.ini', 'redis')

        args, kwargs = list(mocked_render_template.call_args)
        self.assertEquals(args, ('kinto.tpl', 'kinto.ini'))

        redis_url = "redis://localhost:6379"

        self.maxDiff = None  # See the full diff in case of error
        self.assertDictEqual(kwargs, {
            'host': '127.0.0.1',
            'secret': kwargs['secret'],
            'storage_backend': 'kinto_redis.storage',
            'cache_backend': 'kinto_redis.cache',
            'permission_backend': 'kinto_redis.permission',
            'storage_url': redis_url + '/1',
            'cache_url':  redis_url + '/2',
            'permission_url': redis_url + '/3',
            'kinto_version': __version__,
            'config_file_timestamp': strftime('%a, %d %b %Y %H:%M:%S %z')
        })

    @mock.patch('kinto.config.render_template')
    def test_init_memory_values(self, mocked_render_template):
        config.init('kinto.ini', 'memory')

        args, kwargs = list(mocked_render_template.call_args)
        self.assertEquals(args, ('kinto.tpl', 'kinto.ini'))

        self.assertDictEqual(kwargs, {
            'host': '127.0.0.1',
            'secret': kwargs['secret'],
            'storage_backend': 'kinto.core.storage.memory',
            'cache_backend': 'kinto.core.cache.memory',
            'permission_backend': 'kinto.core.permission.memory',
            'storage_url': '',
            'cache_url':  '',
            'permission_url': '',
            'kinto_version': __version__,
            'config_file_timestamp': strftime('%a, %d %b %Y %H:%M:%S %z')
        })

    def test_render_template_works_with_file_in_cwd(self):
        temp_path = tempfile.mkdtemp()
        os.chdir(temp_path)
        config.render_template('kinto.tpl', 'kinto.ini', **{
            'host': '127.0.0.1',
            'secret': "abcd-ceci-est-un-secret",
            'storage_backend': 'kinto.core.storage.memory',
            'cache_backend': 'kinto.core.cache.memory',
            'permission_backend': 'kinto.core.permission.memory',
            'storage_url': '',
            'cache_url':  '',
            'permission_url': '',
            'kinto_version': '',
            'config_file_timestamp': ''
        })
        self.assertTrue(os.path.exists(
            os.path.join(temp_path, 'kinto.ini')
        ))



import mock
import functools

from pyramid.security import IAuthorizationPolicy, Authenticated, Everyone
from zope.interface import implementer

from kinto.core.authorization import PRIVATE
from kinto.core import testing

from .testapp import main as testapp


# This is the principal a connected user should have (in the tests).
USER_PRINCIPAL = 'basicauth:8a931a10fc88ab2f6d1cc02a07d3a81b5d4768f6f13e85c5' \
                 'd8d4180419acb1b4'


class BaseWebTest(testing.BaseWebTest):
    """Base Web Test to test your cornice service.

    It setups the database before each test and delete it after.
    """

    api_prefix = "v0"
    entry_point = testapp
    principal = USER_PRINCIPAL

    authorization_policy = 'tests.core.support.AllowAuthorizationPolicy'
    collection_url = '/mushrooms'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers.update(testing.get_user_headers('mat'))

    def get_app_settings(self, extras=None):
        if extras is None:
            extras = {}
        extras.setdefault('project_name', 'myapp')
        extras.setdefault('project_version', '0.0.1')
        extras.setdefault('http_api_version', '0.1')
        extras.setdefault('project_docs', 'https://kinto.readthedocs.io/')
        extras.setdefault('multiauth.authorization_policy',
                          self.authorization_policy)
        return super().get_app_settings(extras)

    def get_item_url(self, id=None):
        """Return the URL of the item using self.item_url."""
        if id is None:
            id = self.record['id']
        return '{}/{}'.format(self.collection_url, id)


@implementer(IAuthorizationPolicy)
class AllowAuthorizationPolicy:
    def permits(self, context, principals, permission):
        if permission == PRIVATE:
            return Authenticated in principals
        if Everyone in principals:
            return True
        # Kinto-Core default authz policy uses prefixed_userid.
        prefixed = [context.prefixed_userid]
        return USER_PRINCIPAL in (principals + prefixed)

    def principals_allowed_by_permission(self, context, permission):
        raise NotImplementedError()  # PRAGMA NOCOVER


def authorize(permits=True, authz_class=None):
    """Patch the default authorization policy to return what is specified
    in :param:permits.
    """
    if authz_class is None:
        authz_class = 'tests.core.support.AllowAuthorizationPolicy'

    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            with mock.patch(
                    '{}.permits'.format(authz_class),
                    return_value=permits):
                return f(*args, **kwargs)
        return wrapped
    return wrapper



import os
import unittest

from pyramid.config import Configurator

from kinto.core.testing import get_user_headers
from kinto.events import ServerFlushed

from .support import (BaseWebTest,
                      MINIMALIST_BUCKET, MINIMALIST_COLLECTION,
                      MINIMALIST_RECORD)


class FlushViewTest(BaseWebTest, unittest.TestCase):

    collection_url = '/buckets/beers/collections/barley/records'

    def setUp(self):
        super().setUp()

        self.events = []

        bucket = {**MINIMALIST_BUCKET}

        self.alice_headers = {**self.headers, **get_user_headers('alice')}

        resp = self.app.get('/', headers=self.alice_headers)
        alice_principal = resp.json['user']['id']
        bucket['permissions'] = {'write': [alice_principal]}

        # Create shared bucket.
        self.app.put_json('/buckets/beers', bucket,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)

        # Records for alice and bob.
        self.app.post_json(self.collection_url,
                           MINIMALIST_RECORD,
                           headers=self.headers,
                           status=201)
        self.app.post_json(self.collection_url,
                           MINIMALIST_RECORD,
                           headers=self.alice_headers,
                           status=201)

    def tearDown(self):
        self.events = []
        super().tearDown()

    def make_app(self, settings=None, config=None):
        settings = self.get_app_settings(settings)
        config = Configurator(settings=settings)
        config.add_subscriber(self.listener, ServerFlushed)
        config.commit()
        return super().make_app(settings=settings, config=config)

    def get_app_settings(self, extras=None):
        if extras is None:
            extras = {}
        extras.setdefault('flush_endpoint_enabled', True)
        settings = super().get_app_settings(extras)
        return settings

    def listener(self, event):
        self.events.append(event)

    def test_returns_404_if_not_enabled_in_configuration(self):
        extra = {'flush_endpoint_enabled': False}
        app = self.make_app(settings=extra)
        app.post('/__flush__', headers=self.headers, status=404)

    def test_removes_every_records_of_everykind(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.app.get(self.collection_url, headers=self.alice_headers)

        self.app.post('/__flush__', headers=self.headers, status=202)

        self.app.get(self.collection_url, headers=self.headers, status=403)
        self.app.get(self.collection_url,
                     headers=self.alice_headers,
                     status=403)

    def test_event_triggered_post(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.app.get(self.collection_url, headers=self.alice_headers)
        self.app.post('/__flush__', headers=self.headers, status=202)
        self.assertEqual(len(self.events), 1)
        self.assertTrue(isinstance(self.events[0], ServerFlushed))

    def test_can_be_enabled_via_environment(self):
        os.environ['KINTO_FLUSH_ENDPOINT_ENABLED'] = 'true'
        extra = {'flush_endpoint_enabled': False}
        app = self.make_app(settings=extra)
        app.post('/__flush__', headers=self.headers)
        os.environ.pop('KINTO_FLUSH_ENDPOINT_ENABLED')

    def test_flush_returns_json(self):
        response = self.app.post('/__flush__', headers=self.headers, status=202)
        self.assertEquals(response.json, {})



from kinto.core.events import ResourceChanged

from .listener import on_resource_changed


def includeme(config):
    config.add_api_capability('quotas',
                              description='Quotas Management on buckets '
                                          'and collections.',
                              url='https://kinto.readthedocs.io')

    # If StatsD is enabled, monitor execution time of listener.
    listener = on_resource_changed
    if config.registry.statsd:
        key = 'plugins.quotas'
        listener = config.registry.statsd.timer(key)(on_resource_changed)

    # Listen to every resources (except history)
    config.add_subscriber(listener, ResourceChanged,
                          for_resources=('bucket', 'group',
                                         'collection', 'record'))



import random
import string

from kinto.core.storage import generators, exceptions
from pyramid.httpexceptions import HTTPNotFound
from kinto.core.errors import http_error, ERRORS


class NameGenerator(generators.Generator):
    def __call__(self):
        alpha_num = string.ascii_letters + string.digits
        alphabet = alpha_num + '-_'
        letters = [random.SystemRandom().choice(alpha_num)]
        letters += [random.SystemRandom().choice(alphabet) for x in range(7)]

        return ''.join(letters)


class RelaxedUUID(generators.UUID4):
    """A generator that generates UUIDs but accepts any string.
    """
    regexp = generators.Generator.regexp


def object_exists_or_404(request, collection_id, object_id, parent_id=''):
    storage = request.registry.storage
    try:
        return storage.get(collection_id=collection_id,
                           parent_id=parent_id,
                           object_id=object_id)
    except exceptions.RecordNotFoundError:
        # XXX: We gave up putting details about parent id here (See #53).
        details = {
            "id": object_id,
            "resource_name": collection_id
        }
        response = http_error(HTTPNotFound(), errno=ERRORS.MISSING_RESOURCE, details=details)
        raise response



import colander
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.settings import aslist

from kinto.authorization import PERMISSIONS_INHERITANCE_TREE
from kinto.core import utils as core_utils, resource
from kinto.core.storage.memory import extract_record_set


def allowed_from_settings(settings, principals):
    """Returns every permissions allowed from settings for the current user.
    :param settings dict: app settings
    :param principals list: list of principals of current user
    :rtype: dict

    Result example::

        {
            "bucket": {"write", "collection:create"},
            "collection": {"read"}
        }

    XXX: This helper will be useful for Kinto/kinto#894
    """
    perms_settings = {k: aslist(v) for k, v in settings.items()
                      if k.endswith('_principals')}
    from_settings = {}
    for key, allowed_principals in perms_settings.items():
        resource_name, permission, _ = key.split('_')
        # Keep the known permissions only.
        if resource_name not in PERMISSIONS_INHERITANCE_TREE.keys():
            continue
        # Keep the permissions of the current user only.
        if not bool(set(principals) & set(allowed_principals)):
            continue
        # ``collection_create_principals`` means ``collection:create`` in bucket.
        if permission == 'create':
            permission = '{resource_name}:{permission}'.format(
                resource_name=resource_name,
                permission=permission)
            resource_name = {  # resource parents.
                'bucket': '',
                'collection': 'bucket',
                'group': 'bucket',
                'record': 'collection'}[resource_name]
        # Store them in a convenient way.
        from_settings.setdefault(resource_name, set()).add(permission)
    return from_settings


class PermissionsModel:
    id_field = 'id'
    modified_field = 'last_modified'
    deleted_field = 'deleted'

    def __init__(self, request):
        self.request = request

    def timestamp(self, parent_id=None):
        return 0

    def get_records(self, filters=None, sorting=None, pagination_rules=None,
                    limit=None, include_deleted=False, parent_id=None):
        # Invert the permissions inheritance tree.
        perms_descending_tree = {}
        for on_resource, tree in PERMISSIONS_INHERITANCE_TREE.items():
            for obtained_perm, obtained_from in tree.items():
                for from_resource, perms in obtained_from.items():
                    for perm in perms:
                        perms_descending_tree.setdefault(from_resource, {})\
                                             .setdefault(perm, {})\
                                             .setdefault(on_resource, set())\
                                             .add(obtained_perm)

        # Obtain current principals.
        principals = self.request.prefixed_principals

        # Query every possible permission of the current user from backend.
        backend = self.request.registry.permission
        perms_by_object_uri = backend.get_accessible_objects(principals)

        # Check settings for every allowed resources.
        from_settings = allowed_from_settings(self.request.registry.settings, principals)

        # Expand permissions obtained from backend with the object URIs that
        # correspond to permissions allowed from settings.
        allowed_resources = {'bucket', 'collection', 'group'} & set(from_settings.keys())
        if allowed_resources:
            storage = self.request.registry.storage
            every_bucket, _ = storage.get_all(parent_id='', collection_id='bucket')
            for bucket in every_bucket:
                bucket_uri = '/buckets/{id}'.format_map(bucket)
                for res in allowed_resources:
                    resource_perms = from_settings[res]
                    # Bucket is always fetched.
                    if res == 'bucket':
                        perms_by_object_uri.setdefault(bucket_uri, set()).update(resource_perms)
                        continue
                    # Fetch bucket collections and groups.
                    # XXX: wrong approach: query in a loop!
                    every_subobjects, _ = storage.get_all(parent_id=bucket_uri,
                                                          collection_id=res)
                    for subobject in every_subobjects:
                        subobj_uri = bucket_uri + '/{0}s/{1}'.format(res, subobject['id'])
                        perms_by_object_uri.setdefault(subobj_uri, set()).update(resource_perms)

        entries = []
        for object_uri, perms in perms_by_object_uri.items():
            try:
                # Obtain associated res from object URI
                resource_name, matchdict = core_utils.view_lookup(self.request,
                                                                  object_uri)
            except ValueError:
                # Skip permissions entries that are not linked to an object URI
                continue

            # For consistency with event payloads, prefix id with resource name
            matchdict[resource_name + '_id'] = matchdict.get('id')

            # Expand implicit permissions using descending tree.
            permissions = set(perms)
            for perm in perms:
                obtained = perms_descending_tree[resource_name][perm]
                # Related to same resource only and not every sub-objects.
                # (e.g "bucket:write" gives "bucket:read" but not "group:read")
                permissions |= obtained[resource_name]

            entry = dict(uri=object_uri,
                         resource_name=resource_name,
                         permissions=list(permissions),
                         **matchdict)
            entries.append(entry)

        return extract_record_set(entries, filters=filters, sorting=sorting,
                                  pagination_rules=pagination_rules,
                                  limit=limit)


class PermissionsSchema(resource.ResourceSchema):
    uri = colander.SchemaNode(colander.String())
    resource_name = colander.SchemaNode(colander.String())
    permissions = colander.Sequence(colander.SchemaNode(colander.String()))
    bucket_id = colander.SchemaNode(colander.String())
    collection_id = colander.SchemaNode(colander.String(),
                                        missing=colander.drop)
    group_id = colander.SchemaNode(colander.String(),
                                   missing=colander.drop)
    record_id = colander.SchemaNode(colander.String(),
                                    missing=colander.drop)

    class Options:
        preserve_unknown = False


@resource.register(name='permissions',
                   description='List of user permissions',
                   collection_path='/permissions',
                   record_path=None,
                   collection_methods=('GET',),
                   permission=NO_PERMISSION_REQUIRED)
class Permissions(resource.ShareableResource):

    schema = PermissionsSchema

    def __init__(self, request, context=None):
        super().__init__(request, context)
        self.model = PermissionsModel(request)

    def _extract_sorting(self, limit):
        # Permissions entries are not stored with timestamp, so do not
        # force it.
        result = super()._extract_sorting(limit)
        without_last_modified = [s for s in result
                                 if s.field != self.model.modified_field]
        return without_last_modified

    def _extract_filters(self):
        result = super()._extract_filters()
        without_last_modified = [s for s in result
                                 if s.field != self.model.modified_field]
        return without_last_modified



from concurrent.futures import ThreadPoolExecutor, wait

import colander
import transaction
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto import logger
from kinto.core import Service


heartbeat = Service(name="heartbeat", path='/__heartbeat__',
                    description="Server health")


class HeartbeatResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'))


heartbeat_responses = {
    '200': HeartbeatResponseSchema(
        description="Server is working properly."),
    '503': HeartbeatResponseSchema(
        description="One or more subsystems failing.")
}


@heartbeat.get(permission=NO_PERMISSION_REQUIRED, tags=['Utilities'],
               operation_id='__heartbeat__', response_schemas=heartbeat_responses)
def get_heartbeat(request):
    """Return information about server health."""
    status = {}

    def heartbeat_check(name, func):
        status[name] = False
        status[name] = func(request)
        # Since the heartbeat checks run concurrently, their transactions
        # overlap and might end in shared lock errors. By aborting here
        # we clean-up the state on each heartbeat call instead of once at the
        # end of the request. See bug Kinto/kinto#804
        transaction.abort()

    # Start executing heartbeats concurrently.
    heartbeats = request.registry.heartbeats
    pool = ThreadPoolExecutor(max_workers=max(1, len(heartbeats.keys())))
    futures = []
    for name, func in heartbeats.items():
        future = pool.submit(heartbeat_check, name, func)
        future.__heartbeat_name = name  # For logging purposes.
        futures.append(future)

    # Wait for the results, with timeout.
    seconds = float(request.registry.settings['heartbeat_timeout_seconds'])
    done, not_done = wait(futures, timeout=seconds)

    # A heartbeat is supposed to return True or False, and never raise.
    # Just in case, go though results to spot any potential exception.
    for future in done:
        exc = future.exception()
        if exc is not None:
            logger.error("'{}' heartbeat failed.".format(future.__heartbeat_name))
            logger.error(exc)

    # Log timed-out heartbeats.
    for future in not_done:
        name = future.__heartbeat_name
        error_msg = "'{}' heartbeat has exceeded timeout of {} seconds."
        logger.error(error_msg.format(name, seconds))

    # If any has failed, return a 503 error response.
    has_error = not all([v or v is None for v in status.values()])
    if has_error:
        request.response.status = 503

    return status


class LbHeartbeatResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping())


lbheartbeat_responses = {
    '200': LbHeartbeatResponseSchema(
        description="Returned if server is reachable.")
}


lbheartbeat = Service(name="lbheartbeat", path='/__lbheartbeat__',
                      description="Web head health")


@lbheartbeat.get(permission=NO_PERMISSION_REQUIRED, tags=['Utilities'],
                 operation_id='__lbheartbeat__', response_schemas=lbheartbeat_responses)
def get_lbheartbeat(request):
    """Return successful healthy response.

    If the load-balancer tries to access this URL and fails, this means the
    Web head is not operational and should be dropped.
    """
    status = {}
    return status



from kinto.core import testing
from kinto import main as kinto_main
from kinto import DEFAULT_SETTINGS


MINIMALIST_BUCKET = {}
MINIMALIST_COLLECTION = {}
MINIMALIST_GROUP = {'data': dict(members=['fxa:user'])}
MINIMALIST_RECORD = {'data': dict(name="Hulled Barley",
                                  type="Whole Grain")}
USER_PRINCIPAL = 'basicauth:8a931a10fc88ab2f6d1cc02a07d3a81b5d4768f6f13e85c5' \
                 'd8d4180419acb1b4'


class BaseWebTest(testing.BaseWebTest):

    api_prefix = "v1"
    entry_point = kinto_main
    principal = USER_PRINCIPAL

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers.update(testing.get_user_headers('mat'))

    def get_app_settings(self, extras=None):
        settings = {**DEFAULT_SETTINGS}
        if extras is not None:
            settings.update(extras)
        settings = super().get_app_settings(extras=settings)
        return settings

    def create_group(self, bucket_id, group_id, members=None):
        if members is None:
            group = MINIMALIST_GROUP
        else:
            group = {'data': {'members': members}}
        group_url = '/buckets/{}/groups/{}'.format(bucket_id, group_id)
        self.app.put_json(group_url, group,
                          headers=self.headers, status=201)

    def create_bucket(self, bucket_id):
        self.app.put_json('/buckets/{}'.format(bucket_id), MINIMALIST_BUCKET,
                          headers=self.headers, status=201)



from pyramid import httpexceptions
from pyramid.httpexceptions import HTTPTemporaryRedirect
from pyramid.settings import asbool
from pyramid.security import forget, NO_PERMISSION_REQUIRED, Authenticated
from pyramid.view import view_config

from kinto.core.errors import http_error, ERRORS
from kinto.core.logs import logger
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.utils import reapply_cors


@view_config(context=httpexceptions.HTTPForbidden,
             permission=NO_PERMISSION_REQUIRED)
def authorization_required(response, request):
    """Distinguish authentication required (``401 Unauthorized``) from
    not allowed (``403 Forbidden``).
    """
    if Authenticated not in request.effective_principals:
        error_msg = "Please authenticate yourself to use this endpoint."
        response = http_error(httpexceptions.HTTPUnauthorized(),
                              errno=ERRORS.MISSING_AUTH_TOKEN,
                              message=error_msg)
        response.headers.extend(forget(request))
        return response

    if response.content_type != "application/json":
        error_msg = "This user cannot access this resource."
        response = http_error(httpexceptions.HTTPForbidden(),
                              errno=ERRORS.FORBIDDEN,
                              message=error_msg)
    return reapply_cors(request, response)


@view_config(context=httpexceptions.HTTPNotFound,
             permission=NO_PERMISSION_REQUIRED)
def page_not_found(response, request):
    """Return a JSON 404 error response."""
    config_key = 'trailing_slash_redirect_enabled'
    redirect_enabled = request.registry.settings[config_key]
    trailing_slash_redirection_enabled = asbool(redirect_enabled)

    querystring = request.url[(request.url.rindex(request.path) +
                               len(request.path)):]

    errno = ERRORS.MISSING_RESOURCE
    error_msg = "The resource you are looking for could not be found."

    if not request.path.startswith('/{}'.format(request.registry.route_prefix)):
        errno = ERRORS.VERSION_NOT_AVAILABLE
        error_msg = ("The requested API version is not available "
                     "on this server.")
    elif trailing_slash_redirection_enabled:
        redirect = None

        if request.path.endswith('/'):
            path = request.path.rstrip('/')
            redirect = '{}{}'.format(path, querystring)
        elif request.path == '/{}'.format(request.registry.route_prefix):
            # Case for /v0 -> /v0/
            redirect = '/{}/{}'.format(request.registry.route_prefix, querystring)

        if redirect:
            return reapply_cors(request, HTTPTemporaryRedirect(redirect))

    if response.content_type != "application/json":
        response = http_error(httpexceptions.HTTPNotFound(),
                              errno=errno,
                              message=error_msg)
    return reapply_cors(request, response)


@view_config(context=httpexceptions.HTTPServiceUnavailable,
             permission=NO_PERMISSION_REQUIRED)
def service_unavailable(response, request):
    if response.content_type != 'application/json':
        error_msg = ("Service temporary unavailable "
                     "due to overloading or maintenance, please retry later.")
        response = http_error(response, errno=ERRORS.BACKEND,
                              message=error_msg)

    retry_after = request.registry.settings['retry_after_seconds']
    response.headers["Retry-After"] = str(retry_after)
    return reapply_cors(request, response)


@view_config(context=httpexceptions.HTTPMethodNotAllowed,
             permission=NO_PERMISSION_REQUIRED)
def method_not_allowed(context, request):
    if context.content_type == 'application/json':
        return context

    response = http_error(context,
                          errno=ERRORS.METHOD_NOT_ALLOWED,
                          message="Method not allowed on this endpoint.")
    return reapply_cors(request, response)


@view_config(context=Exception, permission=NO_PERMISSION_REQUIRED)
@view_config(context=httpexceptions.HTTPException,
             permission=NO_PERMISSION_REQUIRED)
def error(context, request):
    """Catch server errors and trace them."""
    if isinstance(context, httpexceptions.Response):
        return reapply_cors(request, context)

    if isinstance(context, storage_exceptions.IntegrityError):
        error_msg = "Integrity constraint violated, please retry."
        response = http_error(httpexceptions.HTTPConflict(),
                              errno=ERRORS.CONSTRAINT_VIOLATED,
                              message=error_msg)
        retry_after = request.registry.settings['retry_after_seconds']
        response.headers["Retry-After"] = str(retry_after)
        return reapply_cors(request, response)

    if isinstance(context, storage_exceptions.BackendError):
        logger.critical(context.original, exc_info=True)
        response = httpexceptions.HTTPServiceUnavailable()
        return service_unavailable(response, request)

    logger.error(context, exc_info=True)

    error_msg = "A programmatic error occured, developers have been informed."
    info = request.registry.settings['error_info_link']
    response = http_error(httpexceptions.HTTPInternalServerError(),
                          message=error_msg,
                          info=info)

    return reapply_cors(request, response)



import unittest
import mock

from cornice.service import get_services
from kinto.core.api import OpenAPI

from .support import BaseWebTest


class OpenAPITest(BaseWebTest, unittest.TestCase):

    def setUp(self):
        super(OpenAPITest, self).setUp()
        self.request = mock.MagicMock()
        self.request.registry.settings = self.get_app_settings()
        self.generator = OpenAPI(get_services(), self.request)
        self.api_doc = self.generator.generate()

    def test_assign_base_path(self):
        self.assertEquals(self.api_doc['basePath'], "/{}".format(self.api_prefix))

    def test_default_security_generator(self):
        self.assertEquals(self.api_doc['paths']['/']['get']['security'], [])
        self.assertEquals(self.api_doc['paths']['/mushrooms']['get']['security'],
                          [{'basicauth': []}])

    def test_security_extensions(self):
        method = {
            "type": "oauth2",
            "authorizationUrl": "https://oauth-stable.dev.lcip.org/v1",
            "flow": "implicit",
            "scopes": {"kinto": "Kinto user scope."}
        }

        self.generator.expose_authentication_method("fxa", method)
        api_doc = self.generator.generate()

        self.assertEquals(api_doc['securityDefinitions']['fxa'], method)
        self.assertCountEqual(api_doc['paths']['/mushrooms']['get']['security'],
                              [{'basicauth': []}, {'fxa': ['kinto']}])

    def test_default_tags(self):
        self.assertEquals(self.api_doc['paths']['/mushrooms']['get']['tags'],
                          ['Mushrooms'])
        self.assertEquals(self.api_doc['paths']['/mushrooms/{id}']['get']['tags'],
                          ['Mushrooms'])

    def test_default_operation_ids(self):
        self.assertEquals(self.api_doc['paths']['/mushrooms']['get']['operationId'],
                          'get_mushrooms')
        self.assertEquals(self.api_doc['paths']['/mushrooms/{id}']['get']['operationId'],
                          'get_mushroom')



from bravado_core.request import IncomingRequest, unmarshal_request
from bravado_core.swagger20_validator import ValidationError

from .support import SwaggerTest


class SwaggerRequestsValidationTest(SwaggerTest):

    def setUp(self):
        super().setUp()

        self.request = IncomingRequest()
        self.request.path = {}
        self.request.headers = {}
        self.request.query = {}
        self.request._json = {}
        self.request.json = lambda: self.request._json

    def test_validate_bucket_path(self):
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Buckets'].get_bucket)

    def test_validate_groups_path(self):
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Groups'].get_groups)

    def test_validate_group_path(self):
        paths = [
            {},
            {'bucket_id': 'b1'},
            {'id': 'g1'}
        ]
        for path in paths:
            self.request.path = path
            self.assertRaises(ValidationError, unmarshal_request,
                              self.request, self.resources['Groups'].get_group)

    def test_validate_collections_path(self):
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Collections'].get_collections)

    def test_validate_collection_path(self):
        paths = [
            {},
            {'bucket_id': 'b1'},
            {'id': 'c1'}
        ]
        for path in paths:
            self.request.path = path
            self.assertRaises(ValidationError, unmarshal_request,
                              self.request, self.resources['Collections'].get_collection)

    def test_validate_records_path(self):
        paths = [
            {},
            {'bucket_id': 'b1'},
            {'collection_id': 'c1'}
        ]
        for path in paths:
            self.request.path = path
            self.assertRaises(ValidationError, unmarshal_request,
                              self.request, self.resources['Records'].get_records)

    def test_validate_record_path(self):
        paths = [
            {},
            {'bucket_id': 'b1', 'collection_id': 'c1'},
            {'id': 'r1'},
        ]
        for path in paths:
            self.request.path = path
            self.assertRaises(ValidationError, unmarshal_request,
                              self.request, self.resources['Records'].get_record)

    def test_validate_data(self):
        bodies = [
            {'data': 'aaa'},
        ]
        for body in bodies:
            self.request._json = body
            self.assertRaises(ValidationError, unmarshal_request,
                              self.request, self.resources['Buckets'].create_bucket)

    def test_validate_permissions(self):
        bodies = [
            {'permissions': 'aaa'},
            {'permissions': {'read': 'aaa'}},
            {'permissions': {'read': [111]}},
        ]
        for body in bodies:
            self.request._json = body
            self.assertRaises(ValidationError, unmarshal_request,
                              self.request, self.resources['Buckets'].create_bucket)

    def test_validate_queries(self):
        queries = [
            {'_since': 'aaa'},
            {'_before': 'aaa'},
            {'_limit': 'aaa'},
            {'_token': {}},
        ]
        for query in queries:
            self.request.query = query
            self.assertRaises(ValidationError, unmarshal_request,
                              self.request, self.resources['Buckets'].get_buckets)

    def test_validate_headers(self):
        headers = [
            {'If-None-Match': '123'},
            {'If-Match': '123'},
        ]
        for head in headers:
            self.request.headers = head
            self.assertRaises(ValidationError, unmarshal_request,
                              self.request, self.resources['Buckets'].get_buckets)

    def test_validate_batch_requests_method(self):
        self.request._json = {
            'requests': [
                {
                    'method': 'AAA',
                    'path': '/buckets/b1',
                },
            ]
        }
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Batch'].batch)

    def test_validate_batch_requests_path(self):
        self.request._json = {
            'requests': [
                {
                    'method': 'GET',
                    'path': 123,
                },
            ]
        }
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Batch'].batch)

    def test_validate_batch_requests_body(self):
        self.request._json = {
            'requests': [
                {
                    'method': 'GET',
                    'path': '/buckets/b1',
                    'body': []
                },
            ]
        }
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Batch'].batch)

    def test_validate_batch_requests_header(self):
        self.request._json = {
            'requests': [
                {
                    'method': 'GET',
                    'path': '/buckets/b1',
                    'body': {},
                    'headers': []
                },
            ]
        }
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Batch'].batch)

    def test_validate_batch_defaults(self):
        self.request._json = {
            'defaults': [],
            'requests': [
                {
                    'method': 'GET',
                    'path': '/buckets/b1',
                },
            ]
        }
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Batch'].batch)

    def test_validate_batch_defaults_method(self):
        self.request._json = {
            'defaults': {
                'method': 'AAA'
            },
            'requests': [
                {
                    'path': '/buckets/b1',
                },
            ]
        }
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Batch'].batch)

    def test_validate_batch_defaults_body(self):
        self.request._json = {
            'defaults': {
                'body': []
            },
            'requests': [
                {
                    'method': 'PUT',
                    'path': '/buckets/b1',
                },
            ]
        }
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Batch'].batch)

    def test_validate_batch_defaults_headers(self):
        self.request._json = {
            'defaults': {
                'headers': []
            },
            'requests': [
                {
                    'method': 'GET',
                    'path': '/buckets/b1',
                },
            ]
        }
        self.assertRaises(ValidationError, unmarshal_request,
                          self.request, self.resources['Batch'].batch)



import os.path
from urllib.parse import urljoin

import random
import re
import unittest
import uuid

import requests


__HERE__ = os.path.abspath(os.path.dirname(__file__))

SERVER_URL = "http://localhost:8888/v1"
DEFAULT_AUTH = ('user', 'p4ssw0rd')


def build_task():
    suffix = str(uuid.uuid4())
    data = {
        "description": "Task description {0}".format(suffix),
        "status": random.choice(("todo", "doing")),
    }
    return data


class FunctionalTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(FunctionalTest, self).__init__(*args, **kwargs)
        # XXX Read the configuration from env variables.
        self.server_url = SERVER_URL
        self.auth = DEFAULT_AUTH
        self.session = requests.Session()
        self.session.auth = DEFAULT_AUTH

    def tearDown(self):
        # Delete all the created objects
        flush_url = urljoin(self.server_url, '/__flush__')
        resp = requests.post(flush_url)
        resp.raise_for_status()

    def test_user_default_bucket_tutorial(self):
        collection_id = 'tasks-%s' % uuid.uuid4()
        collection_url = urljoin(self.server_url,
                                 '/buckets/default/collections/{}/records'.format(collection_id))
        task = build_task()

        # Create a task on a default collection
        # 201 created with data and permission.write
        resp = self.session.post(collection_url, json={"data": task})
        resp.raise_for_status()
        self.assertEquals(resp.status_code, 201)
        record = resp.json()

        self.assertEquals(record['data']['description'], task['description'])
        self.assertEquals(record['data']['status'], task['status'])
        self.assertIn('write', record['permissions'])

        # Create a new one with PUT and If-None-Match: "*"
        task = build_task()
        record_id = str(uuid.uuid4())
        record_url = urljoin(self.server_url,
                             '{}/{}'.format(collection_url, record_id))
        resp = self.session.put(record_url, json={'data': task},
                                headers={'If-None-Match': '*'})
        resp.raise_for_status()
        resp.raise_for_status()
        self.assertEquals(resp.status_code, 201)
        record = resp.json()

        self.assertEquals(record['data']['description'], task['description'])
        self.assertEquals(record['data']['status'], task['status'])
        self.assertIn('write', record['permissions'])

        # Fetch the collection list and see the tasks (save the etag)
        resp = self.session.get(collection_url)
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(len(record['data']), 2)
        etag = resp.headers['ETag']

        # Fetch the collection from the Etag and see nothing new
        resp = self.session.get(
            collection_url,
            params={'_since': etag.strip('"')})
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(len(record['data']), 0)

        # Update a task
        resp = self.session.patch(
            record_url,
            json={'data': {'status': 'done'}})
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(record['data']['status'], 'done')
        self.assertIn('write', record['permissions'])

        # Try an update with If-Match on the saved ETag and see it fails
        resp = self.session.patch(
            record_url,
            json={'data': {'status': 'done'}},
            headers={'If-Match': etag})
        self.assertEqual(resp.status_code, 412)
        self.assertEqual(resp.headers['ETag'],
                         '"%s"' % record['data']['last_modified'])

        # Get the list of records and update the ETag
        resp = self.session.get(
            '%s?_since=%s' % (collection_url, etag.strip('"')))
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(len(record['data']), 1)
        etag = resp.headers['ETag']

        # Try an update with If-Match on the new ETag and see it works
        resp = self.session.patch(
            record_url,
            json={'data': {'status': 'done'}},
            headers={'If-Match': etag})
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        etag = resp.headers['ETag']

        # Delete the record with If-Match
        resp = self.session.delete(
            record_url,
            headers={'If-Match': etag})

        self.assertEqual(resp.status_code, 200)

        # Try the collection get with the ``_since`` parameter
        resp = self.session.get(
            '%s?_since=%s' % (collection_url, etag.strip('"')),
            headers={'If-None-Match': etag})
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertEqual(len(record['data']), 1)
        self.assertIn('deleted', record['data'][0])

        # Delete all the things
        resp = self.session.delete(collection_url)
        self.assertEqual(resp.status_code, 200)

    def test_user_shared_bucket_tutorial(self):
        bucket_id = 'bucket-%s' % uuid.uuid4()
        collection_id = 'tasks-%s' % uuid.uuid4()
        bucket_url = urljoin(self.server_url, '/buckets/{}'.format(bucket_id))
        collection_url = '{}/collections/{}/records'.format(bucket_url, collection_id)

        # Create a new bucket and check for permissions
        resp = self.session.put(bucket_url)
        # In case of concurrent execution, it may have been created already.
        self.assertIn(resp.status_code, (200, 201))
        record = resp.json()
        self.assertIn('write', record['permissions'])

        # Create a new collection and check for permissions
        permissions = {"record:create": ['system.Authenticated']}
        resp = self.session.put(
            re.sub('/records$', '', collection_url),
            json={'permissions': permissions})
        # In case of concurrent execution, it may have been created already.
        self.assertIn(resp.status_code, (200, 201))
        record = resp.json()
        self.assertIn('record:create', record['permissions'])
        self.assertIn('system.Authenticated',
                      record['permissions']['record:create'])

        # Create a new tasks for Alice
        alice_auth = ('token', 'alice-secret-%s' % uuid.uuid4())
        alice_task = build_task()
        resp = self.session.post(
            collection_url,
            json={'data': alice_task},
            auth=alice_auth)
        self.assertEqual(resp.status_code, 201)
        record = resp.json()
        self.assertIn('write', record['permissions'])
        alice_task_id = record['data']['id']

        bob_auth = ('token', 'bob-secret-%s' % uuid.uuid4())

        # Bob has no task yet.
        resp = self.session.get(collection_url, auth=bob_auth)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()['data']), 0)

        # Create a new tasks for Bob
        bob_task = build_task()
        resp = self.session.post(
            collection_url,
            json={'data': bob_task},
            auth=bob_auth)
        self.assertEqual(resp.status_code, 201)
        record = resp.json()
        self.assertIn('write', record['permissions'])
        bob_user_id = record['permissions']['write'][0]
        bob_task_id = record['data']['id']

        # Now that he has a task, he should see his.
        resp = self.session.get(collection_url, auth=bob_auth)
        self.assertEqual(resp.status_code, 200)
        records = resp.json()['data']
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]['id'], bob_task_id)

        record_url = '{}/{}'.format(collection_url, alice_task_id)

        # Share Alice's task with Bob
        resp = self.session.patch(record_url,
                                  json={'permissions': {'read': [bob_user_id]}},
                                  auth=alice_auth)
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertIn('write', record['permissions'])
        alice_task_id = record['data']['id']

        # Check that Bob can access it
        resp = self.session.get(record_url, auth=bob_auth)
        self.assertEqual(resp.status_code, 200)

        # Get mary's userid
        mary_auth = ('token', 'mary-secret-%s' % uuid.uuid4())
        resp = self.session.get('{}/'.format(self.server_url), auth=mary_auth)
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        mary_user_id = record['user']['id']

        # Allow group creation on bucket
        permissions = {"group:create": ['system.Authenticated']}
        resp = self.session.put(bucket_url,
                                json={'permissions': permissions})
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertIn('group:create', record['permissions'])
        self.assertIn('system.Authenticated',
                      record['permissions']['group:create'])

        # Create Alice's friend group with Bob and Mary
        group_url = '{}/groups/alices-friends'.format(bucket_url)
        resp = self.session.put(group_url,
                                json={'data': {'members': [mary_user_id, bob_user_id]}},
                                auth=alice_auth)
        self.assertEqual(resp.status_code, 201)

        # Give Alice's task permission for that group
        group_id = '/buckets/{}/groups/alices-friends'.format(bucket_id)
        resp = self.session.patch(record_url,
                                  json={'permissions': {'read': [group_id]}},
                                  auth=alice_auth)
        self.assertEqual(resp.status_code, 200)
        record = resp.json()
        self.assertIn(group_id, record['permissions']['read'])

        # Try to access Alice's task with Mary
        resp = self.session.get(record_url, auth=mary_auth)
        self.assertEqual(resp.status_code, 200)

        # Check that Mary's collection_get sees Alice's task
        resp = self.session.get(collection_url, auth=mary_auth)
        self.assertEqual(resp.status_code, 200)
        records = resp.json()['data']
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]['id'], alice_task_id)

        # Check that Bob's collection_get sees both his and Alice's tasks
        resp = self.session.get(collection_url, auth=bob_auth)
        self.assertEqual(resp.status_code, 200)
        records = resp.json()['data']
        self.assertEqual(len(records), 2)
        records_ids = [r['id'] for r in records]
        self.assertIn(alice_task_id, records_ids)
        self.assertIn(bob_task_id, records_ids)

    def test_check_for_lists(self):
        # List buckets should not be forbidden
        resp = self.session.get('{}/buckets'.format(self.server_url))
        self.assertEqual(resp.status_code, 200)



from kinto.core import resource
from kinto.core.utils import instance_uri
from kinto.core.events import ResourceChanged, ACTIONS
from pyramid.events import subscriber


@resource.register(name='bucket',
                   collection_path='/buckets',
                   record_path='/buckets/{{id}}')
class Bucket(resource.ShareableResource):
    permissions = ('read', 'write', 'collection:create', 'group:create')

    def get_parent_id(self, request):
        # Buckets are not isolated by user, unlike Kinto-Core resources.
        return ''


@subscriber(ResourceChanged,
            for_resources=('bucket',),
            for_actions=(ACTIONS.DELETE,))
def on_buckets_deleted(event):
    """Some buckets were deleted, delete sub-resources.
    """
    storage = event.request.registry.storage
    permission = event.request.registry.permission

    for change in event.impacted_records:
        bucket = change['old']
        bucket_uri = instance_uri(event.request, 'bucket', id=bucket['id'])
        # Delete everything whose parent_id starts with bucket_uri.
        parent_pattern = bucket_uri + '*'
        storage.delete_all(parent_id=parent_pattern,
                           collection_id=None,
                           with_deleted=False)
        # Remove remaining tombstones too.
        storage.purge_deleted(parent_id=parent_pattern,
                              collection_id=None)
        # Remove related permissions
        permission.delete_object_permissions(parent_pattern)



from pyramid.static import static_view
from pyramid.httpexceptions import HTTPTemporaryRedirect
from .views import admin_home_view


def includeme(config):
    # Process settings to remove storage wording.

    # Expose capability.
    config.add_api_capability(
        "admin",
        version="1.13.3",
        description="Serves the admin console.",
        url="https://github.com/Kinto/kinto-admin/",
    )

    config.add_route('admin_home', '/admin/')
    config.add_view(admin_home_view, route_name="admin_home")

    build_dir = static_view('kinto.plugins.admin:build', use_subpath=True)
    config.add_route('catchall_static', '/admin/*subpath')
    config.add_view(build_dir, route_name="catchall_static")

    # Setup redirect without trailing slash.
    def admin_redirect_view(request):
        raise HTTPTemporaryRedirect(request.path + '/')

    config.add_route('admin_redirect', '/admin')
    config.add_view(admin_redirect_view, route_name="admin_redirect")



import os
import codecs
from time import strftime

from kinto.core import utils as core_utils

from kinto import logger
from kinto import __version__

HERE = os.path.abspath(os.path.dirname(__file__))


def render_template(template, destination, **kwargs):
    template = os.path.join(HERE, template)
    folder = os.path.dirname(destination)

    if folder and not os.path.exists(folder):
        os.makedirs(folder)

    logger.info("Created config {}".format(os.path.abspath(destination)))

    with codecs.open(template, 'r', encoding='utf-8') as f:
        raw_template = f.read()
        rendered = raw_template.format_map(kwargs)
        with codecs.open(destination, 'w+', encoding='utf-8') as output:
            output.write(rendered)


def init(config_file, backend, host='127.0.0.1'):
    values = {}

    values['host'] = host
    values['secret'] = core_utils.random_bytes_hex(32)

    values['kinto_version'] = __version__
    values['config_file_timestamp'] = str(strftime('%a, %d %b %Y %H:%M:%S %z'))

    values['storage_backend'] = "kinto.core.storage.{}".format(backend)
    values['cache_backend'] = "kinto.core.cache.{}".format(backend)
    values['permission_backend'] = "kinto.core.permission.{}".format(backend)

    if backend == 'postgresql':
        postgresql_url = "postgres://postgres:postgres@localhost/postgres"
        values['storage_url'] = postgresql_url
        values['cache_url'] = postgresql_url
        values['permission_url'] = postgresql_url

    elif backend == 'redis':
        redis_url = "redis://localhost:6379"
        values['storage_backend'] = "kinto_redis.storage"
        values['cache_backend'] = "kinto_redis.cache"
        values['permission_backend'] = "kinto_redis.permission"

        values['storage_url'] = redis_url + "/1"
        values['cache_url'] = redis_url + "/2"
        values['permission_url'] = redis_url + "/3"

    else:
        values['storage_url'] = ''
        values['cache_url'] = ''
        values['permission_url'] = ''

    render_template("kinto.tpl", config_file, **values)



"""
kinto.core.scripts: utilities to build admin scripts for kinto-based services
"""

import transaction as current_transaction
from pyramid.settings import asbool

from kinto import logger
from kinto.core.storage import exceptions as storage_exceptions


def migrate(env, dry_run=False):
    """
    User-friendly frontend to run database migrations.
    """
    registry = env['registry']
    settings = registry.settings
    readonly_backends = ('storage', 'permission')
    readonly_mode = asbool(settings.get('readonly', False))

    for backend in ('cache', 'storage', 'permission'):
        if hasattr(registry, backend):
            if readonly_mode and backend in readonly_backends:
                message = ('Cannot migrate the {} backend while '
                           'in readonly mode.'.format(backend))
                logger.error(message)
            else:
                getattr(registry, backend).initialize_schema(dry_run=dry_run)


def delete_collection(env, bucket_id, collection_id):
    registry = env['registry']
    settings = registry.settings
    readonly_mode = asbool(settings.get('readonly', False))

    if readonly_mode:
        message = ('Cannot delete the collection while in readonly mode.')
        logger.error(message)
        return 31

    bucket = '/buckets/{}'.format(bucket_id)
    collection = '/buckets/{}/collections/{}'.format(bucket_id, collection_id)

    try:
        registry.storage.get(collection_id='bucket',
                             parent_id='',
                             object_id=bucket_id)
    except storage_exceptions.RecordNotFoundError:
        logger.error("Bucket '{}' does not exist.".format(bucket))
        return 32

    try:
        registry.storage.get(collection_id='collection',
                             parent_id=bucket,
                             object_id=collection_id)
    except storage_exceptions.RecordNotFoundError:
        logger.error("Collection '{}' does not exist.".format(collection))
        return 33

    deleted = registry.storage.delete_all(collection_id='record',
                                          parent_id=collection,
                                          with_deleted=False)
    if len(deleted) == 0:
        logger.info("No records found for '{}'.".format(collection))
    else:
        logger.info('{} record(s) were deleted.'.format(len(deleted)))

    registry.storage.delete(collection_id='collection',
                            parent_id=bucket,
                            object_id=collection_id,
                            with_deleted=False)
    logger.info("'{}' collection object was deleted.".format(collection))

    record = ('/buckets/{bucket_id}'
              '/collections/{collection_id}'
              '/records/{record_id}')

    registry.permission.delete_object_permissions(
        collection,
        *[record.format(bucket_id=bucket_id,
                        collection_id=collection_id,
                        record_id=r['id']) for r in deleted])
    logger.info('Related permissions were deleted.')

    current_transaction.commit()

    return 0



import threading
from functools import update_wrapper
from pyramid.response import Response


class cache_forever:
    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.saved = None
        update_wrapper(self, wrapped)

    def __call__(self, request, *args, **kwargs):
        if self.saved is None:
            self.saved = self.wrapped(request, *args, **kwargs)
            if isinstance(self.saved, Response):
                self.saved = None
                raise ValueError('cache_forever cannot cache Response only its body')

        request.response.write(self.saved)
        return request.response


def synchronized(method):
    """Class method decorator to make sure two threads do not execute some code
    at the same time (c.f Java ``synchronized`` keyword).

    The decorator installs a mutex on the class instance.
    """
    def decorated(self, *args, **kwargs):
        try:
            lock = getattr(self, '__lock__')
        except AttributeError:
            lock = threading.RLock()
            setattr(self, '__lock__', lock)

        lock.acquire()
        try:
            result = method(self, *args, **kwargs)
        finally:
            lock.release()
        return result
    return decorated



import unittest
import json

from bravado_core.spec import Spec
from bravado_core.resource import build_resources
from bravado_core.response import OutgoingResponse, validate_response
from bravado_core.request import IncomingRequest, unmarshal_request

from ..support import (BaseWebTest, MINIMALIST_BUCKET, MINIMALIST_GROUP,
                       MINIMALIST_COLLECTION, MINIMALIST_RECORD)


class SwaggerTest(BaseWebTest, unittest.TestCase):

    settings = {
        'includes': [
            'kinto.plugins.history',
            'kinto.plugins.admin',
        ]
    }

    @classmethod
    def setUpClass(cls):
        # FIXME: solve memory issues from generating the spec multiple times
        app = BaseWebTest().make_app(settings=cls.settings)

        cls.spec_dict = app.get('/__api__').json
        cls.spec = Spec.from_dict(cls.spec_dict)
        cls.resources = build_resources(cls.spec)

    def setUp(self):
        super().setUp()

        self.bucket = self.app.put_json('/buckets/b1',
                                        MINIMALIST_BUCKET,
                                        headers=self.headers).json

        self.group = self.app.put_json('/buckets/b1/groups/g1',
                                       MINIMALIST_GROUP,
                                       headers=self.headers).json

        self.collection = self.app.put_json('/buckets/b1/collections/c1',
                                            MINIMALIST_COLLECTION,
                                            headers=self.headers).json

        self.record = self.app.put_json('/buckets/b1/collections/c1/records/r1',
                                        MINIMALIST_RECORD,
                                        headers=self.headers).json

        # Create raw Bravado request
        self.request = IncomingRequest()
        self.request.url = ''
        self.request.data = ''
        self.request.path = {}
        self.request.headers = {}
        self.request.query = {}
        self.request._json = {}
        self.request.json = lambda: self.request._json

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings.update(self.settings)
        return settings

    def cast_bravado_response(self, response):
        resp = OutgoingResponse()
        resp.text = response.body

        # FIXME: Empty response fails with Bravado and Python 3
        # See https://github.com/Yelp/bravado-core/issues/134
        if not resp.text:
            resp.text = None

        resp.headers = response.headers
        # Response headers integer fields are not casted by default
        for k, v in resp.headers.items():
            try:
                resp.headers[k] = int(v)
            except ValueError:
                pass

        # FIXME: Drop charset from application/json response on Pyramid
        # See https://github.com/Pylons/pyramid/issues/2860
        resp.content_type = response.headers.get('Content-Type', '').split(';')[0]
        resp.json = lambda: response.json

        return resp

    def validate_request_call(self, op, **kargs):
        params = unmarshal_request(self.request, op)
        response = self.app.request(op.path_name.format_map(params),
                                    body=json.dumps(self.request.json()).encode(),
                                    method=op.http_method.upper(),
                                    headers=self.headers, **kargs)
        schema = self.spec.deref(op.op_spec['responses'][str(response.status_code)])
        casted_resp = self.cast_bravado_response(response)
        validate_response(schema, op, casted_resp)
        return response



from bravado_core.response import validate_response

from .support import SwaggerTest, MINIMALIST_BUCKET


class SwaggerBucketResponsesTest(SwaggerTest):

    def test_get_bucket_200(self):
        response = self.app.get('/buckets/b1',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_bucket
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_post_bucket_200(self):
        response = self.app.post_json('/buckets', self.bucket,
                                      headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].create_bucket
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_post_bucket_201(self):
        response = self.app.post_json('/buckets', MINIMALIST_BUCKET,
                                      headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].create_bucket
        schema = self.spec.deref(op.op_spec['responses']['201'])
        validate_response(schema, op, response)

    def test_put_bucket_200(self):
        response = self.app.put('/buckets/b1',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].update_bucket
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_put_bucket_201(self):
        response = self.app.put('/buckets/b2',
                                headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].update_bucket
        schema = self.spec.deref(op.op_spec['responses']['201'])
        validate_response(schema, op, response)

    def test_delete_bucket_200(self):
        response = self.app.delete('/buckets/b1',
                                   headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_bucket
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_get_buckets_200(self):
        response = self.app.get('/buckets',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].get_buckets
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_delete_buckets_200(self):
        response = self.app.delete('/buckets',
                                   headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Buckets'].delete_buckets
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)



import argparse
import os
import sys
import logging
import logging.config

from kinto.core import scripts
from pyramid.scripts import pserve
from pyramid.paster import bootstrap
from kinto import __version__
from kinto.config import init

DEFAULT_CONFIG_FILE = 'config/kinto.ini'
DEFAULT_PORT = 8888
DEFAULT_LOG_LEVEL = logging.INFO
DEFAULT_LOG_FORMAT = "%(levelname)-5.5s  %(message)s"


def main(args=None):
    """The main routine."""
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(description="Kinto Command-Line "
                                                 "Interface")
    commands = ('init', 'start', 'migrate', 'delete-collection', 'version')
    subparsers = parser.add_subparsers(title='subcommands',
                                       description='Main Kinto CLI commands',
                                       dest='subcommand',
                                       help="Choose and run with --help")
    subparsers.required = True

    for command in commands:
        subparser = subparsers.add_parser(command)
        subparser.set_defaults(which=command)

        subparser.add_argument('--ini',
                               help='Application configuration file',
                               dest='ini_file',
                               required=False,
                               default=DEFAULT_CONFIG_FILE)

        subparser.add_argument('-q', '--quiet', action='store_const',
                               const=logging.CRITICAL, dest='verbosity',
                               help='Show only critical errors.')

        subparser.add_argument('-v', '--debug', action='store_const',
                               const=logging.DEBUG, dest='verbosity',
                               help='Show all messages, including debug messages.')

        if command == 'init':
            subparser.add_argument('--backend',
                                   help='{memory,redis,postgresql}',
                                   dest='backend',
                                   required=False,
                                   default=None)
            subparser.add_argument('--host',
                                   help='Host to listen() on.',
                                   dest='host',
                                   required=False,
                                   default='127.0.0.1')
        elif command == 'migrate':
            subparser.add_argument('--dry-run',
                                   action='store_true',
                                   help='Simulate the migration operations '
                                        'and show information',
                                   dest='dry_run',
                                   required=False,
                                   default=False)
        elif command == 'delete-collection':
            subparser.add_argument('--bucket',
                                   help='The bucket where the collection '
                                        'belongs to.',
                                   required=True)
            subparser.add_argument('--collection',
                                   help='The collection to remove.',
                                   required=True)

        elif command == 'start':
            subparser.add_argument('--reload',
                                   action='store_true',
                                   help='Restart when code or config changes',
                                   required=False,
                                   default=False)
            subparser.add_argument('--port',
                                   type=int,
                                   help='Listening port number',
                                   required=False,
                                   default=DEFAULT_PORT)

    # Parse command-line arguments
    parsed_args = vars(parser.parse_args(args))

    config_file = parsed_args['ini_file']
    which_command = parsed_args['which']

    # Initialize logging from
    level = parsed_args.get('verbosity') or DEFAULT_LOG_LEVEL
    logging.basicConfig(level=level, format=DEFAULT_LOG_FORMAT)

    if which_command == 'init':
        if os.path.exists(config_file):
            print("{} already exists.".format(config_file), file=sys.stderr)
            return 1

        backend = parsed_args['backend']
        if not backend:
            while True:
                prompt = ("Select the backend you would like to use: "
                          "(1 - postgresql, 2 - redis, default - memory) ")
                answer = input(prompt).strip()
                try:
                    backends = {"1": "postgresql", "2": "redis", "": "memory"}
                    backend = backends[answer]
                    break
                except KeyError:
                    pass

        init(config_file, backend, parsed_args['host'])

        # Install postgresql libraries if necessary
        if backend == "postgresql":
            try:
                import psycopg2  # NOQA
            except ImportError:
                import pip
                pip.main(['install', "kinto[postgresql]"])
        elif backend == "redis":
            try:
                import kinto_redis  # NOQA
            except ImportError:
                import pip
                pip.main(['install', "kinto[redis]"])

    elif which_command == 'migrate':
        dry_run = parsed_args['dry_run']
        env = bootstrap(config_file)
        scripts.migrate(env, dry_run=dry_run)

    elif which_command == 'delete-collection':
        env = bootstrap(config_file)
        return scripts.delete_collection(env,
                                         parsed_args['bucket'],
                                         parsed_args['collection'])

    elif which_command == 'start':
        pserve_argv = ['pserve']

        if parsed_args['reload']:
            pserve_argv.append('--reload')

        if level == logging.DEBUG:
            pserve_argv.append('-v')

        if level == logging.CRITICAL:
            pserve_argv.append('-q')

        pserve_argv.append(config_file)
        pserve_argv.append('http_port={}'.format(parsed_args['port']))
        pserve.main(argv=pserve_argv)

    else:
        print(__version__)

    return 0



import json

from kinto.core import logger
from kinto.core.listeners import ListenerBase


class Listener(ListenerBase):
    """
    A demo listener that just log received info in the stdout.
    """
    def __call__(self, event):
        try:
            logger.info(json.dumps(event.payload))
        except TypeError:
            logger.error("Unable to dump the payload", exc_info=True)
            return


def load_from_config(config, prefix):
    return Listener()
